from builtins import range
import mdp.parallel as parallel
from ._tools import *

requires_parallel_python = skip_on_condition(
    "not mdp.config.has_parallel_python",
    "This test requires Parallel Python")


@requires_parallel_python
def test_reverse_patching():
    # revert pp patching
    # XXX This is needed to avoid failures of the other
    # XXX pp tests when run more then once in the same interpreter
    # XXX session
    if hasattr(mdp.config, 'pp_monkeypatch_dirname'):
        import pp
        pp._Worker.command = mdp._pp_worker_command[:]
        parallel.pp_support._monkeypatch_pp(mdp.config.pp_monkeypatch_dirname)


@requires_parallel_python
def test_simple():
    """Test local pp scheduling."""
    scheduler = parallel.pp_support.LocalPPScheduler(ncpus=2,
                                                     max_queue_length=0,
                                                     verbose=False)
    # process jobs
    for i in range(50):
        scheduler.add_task(i, parallel.SqrTestCallable())
    results = scheduler.get_results()
    scheduler.shutdown()
    # check result
    results.sort()
    results = numx.array(results[:6])
    assert numx.all(results == numx.array([0,1,4,9,16,25]))

@requires_parallel_python
def test_scheduler_flow():
    """Test local pp scheduler with real Nodes."""
    precision = 10**-6
    node1 = mdp.nodes.PCANode(output_dim=20)
    node2 = mdp.nodes.PolynomialExpansionNode(degree=1)
    node3 = mdp.nodes.SFANode(output_dim=10)
    flow = mdp.parallel.ParallelFlow([node1, node2, node3])
    parallel_flow = mdp.parallel.ParallelFlow(flow.copy()[:])
    scheduler = parallel.pp_support.LocalPPScheduler(ncpus=3,
                                                     max_queue_length=0,
                                                     verbose=False)
    input_dim = 30
    scales = numx.linspace(1, 100, num=input_dim)
    scale_matrix = mdp.numx.diag(scales)
    train_iterables = [numx.dot(mdp.numx_rand.random((5, 100, input_dim)),
                             scale_matrix)
                       for _ in range(3)]
    parallel_flow.train(train_iterables, scheduler=scheduler)
    x = mdp.numx.random.random((10, input_dim))
    # test that parallel execution works as well
    # note that we need more chungs then processes to test caching
    parallel_flow.execute([x for _ in range(8)], scheduler=scheduler)
    scheduler.shutdown()
    # compare to normal flow
    flow.train(train_iterables)
    assert parallel_flow[0].tlen == flow[0].tlen
    y1 = flow.execute(x)
    y2 = parallel_flow.execute(x)
    assert_array_almost_equal(abs(y1 - y2), precision)



from builtins import range
from ._tools import *

def hardcoded_expansion(x, degree):
    nvars = x.shape[1]
    exp_dim = mdp.nodes._expanded_dim(degree, nvars)
    exp = numx.zeros((x.shape[0], exp_dim), 'd')
    # degree 1
    exp[:,:nvars] = x.copy()
    # degree 2
    k = nvars
    if degree>=2:
        for i in range(nvars):
            for j in range(i,nvars):
                exp[:,k] = x[:,i]*x[:,j]
                k += 1
    # degree 3
    if degree>=3:
        for i in range(nvars):
            for j in range(i,nvars):
                for l in range(j,nvars):
                    exp[:,k] = x[:,i]*x[:,j]*x[:,l]
                    k += 1
    # degree 4
    if degree>=4:
        for i in range(nvars):
            for j in range(i,nvars):
                for l in range(j,nvars):
                    for m in range(l,nvars):
                        exp[:,k] = x[:,i]*x[:,j]*x[:,l]*x[:,m]
                        k += 1
    # degree 5
    if degree>=5:
        for i in range(nvars):
            for j in range(i,nvars):
                for l in range(j,nvars):
                    for m in range(l,nvars):
                        for n in range(m,nvars):
                            exp[:,k] = \
                                     x[:,i]*x[:,j]*x[:,l]*x[:,m]*x[:,n]
                            k += 1
    return exp

def test_expansion():
    for degree in range(1,6):
        for dim in range(1,5):
            expand = mdp.nodes.PolynomialExpansionNode(degree=degree)
            mat,mix,inp = get_random_mix((10,dim))
            des = hardcoded_expansion(inp, degree)
            exp = expand.execute(inp)
            assert_array_almost_equal(exp, des, decimal)



"""
Extension Mechanism for nodes.

The extension mechanism makes it possible to dynamically add class attributes,
especially methods, for specific features to node classes
(e.g. nodes need a _fork and _join method for parallelization).
It is also possible for users to define new extensions to provide new
functionality for MDP nodes without having to modify any MDP code.

Without the extension mechanism extending nodes would be done by inheritance,
which is fine unless one wants to use multiple inheritance at the same time
(requiring multiple inheritance for every combination of extensions one wants
to use). The extension mechanism does not depend on inheritance, instead it
adds the methods to the node classes dynamically at runtime. This makes it
possible to activate extensions just when they are needed, reducing the risk
of interference between different extensions.

However, since the extension mechanism provides a special Metaclass it is
still possible to define the extension nodes as classes derived from nodes.
This keeps the code readable and is compatible with automatic code checkers
(like the background pylint checks in the Eclipse IDE with PyDev).
"""
from __future__ import print_function
from builtins import str
from builtins import object

from mdp import MDPException, NodeMetaclass
from future.utils import with_metaclass

# TODO: Register the node instances as well?
#    This would allow instance initialization when an extension is activated.
#    Implementing this should not be too hard via the metclass.

# TODO: Add warning about overriding public methods with respect to
#    the docstring wrappers?

# TODO: in the future could use ABC's to register nodes with extension nodes


# name prefix used for the original attributes when they are shadowed
ORIGINAL_ATTR_PREFIX = "_non_extension_"
# prefix used to store the current extension name for an attribute,
# the value stored in this attribute is the extension name
_EXTENSION_ATTR_PREFIX = "_extension_for_"
# list of attribute names that are not affected by extensions,
_NON_EXTENSION_ATTRIBUTES = ["__module__", "__doc__", "extension_name"]

# keys under which the global activation and deactivation functions
# for extensions can be stored in the extension registry
_SETUP_FUNC_ATTR = "_extension_setup"
_TEARDOWN_FUNC_ATTR = "_extension_teardown"

# dict of dicts of dicts, contains a key for each extension,
# the inner dict maps the node types to their extension node,
# the innermost dict then maps attribute names to values
# (e.g. a method name to the actual function)
#
# For each extension there are also the special _SETUP_FUNC_ATTR and
# _TEARDOWN_FUNC_ATTR keys.
_extensions = dict()
# set containing the names of the currently activated extensions
_active_extensions = set()


class ExtensionException(MDPException):
    """Base class for extension related exceptions."""
    pass


def _register_attribute(ext_name, node_cls, attr_name, attr_value):
    """Register an attribute as an extension attribute.

    ext_name -- String with the name of the extension.
    node_cls -- Node class for which the method should be registered.
    """
    _extensions[ext_name][node_cls][attr_name] = attr_value


def extension_method(extension_name, node_cls, method_name=None):
    """Returns a decorator to register a function as an extension method.

    :Parameters:
      extension_name
        String with the name of the extension.
      node_cls
        Node class for which the method should be registered.
      method_name
        Name of the extension method (default value is ``None``).

        If no value is provided then the name of the function is used.

    Note that it is possible to directly call other extension functions, call
    extension methods in other node classes or to use super in the normal way
    (the function will be called as a method of the node class).
    """
    def register_function(func):
        _method_name = method_name
        if not _method_name:
            _method_name = func.__name__
        if not extension_name in _extensions:
            # creation of a new extension, add entry in dict
            _extensions[extension_name] = dict()
        if not node_cls in _extensions[extension_name]:
            # register this node
            _extensions[extension_name][node_cls] = dict()
        _register_attribute(extension_name, node_cls, _method_name, func)
        return func
    return register_function


def extension_setup(extension_name):
    """Returns a decorator to register a setup function for an extension.

    :Parameters:
      extension_name
        String with the name of the extension.

    The decorated function will be called when the extension is activated.
    
    Note that there is also the extension_teardown decorator, which should
    probably defined as well if there is a setup procedure.
    """
    def register_setup_function(func):
        if not extension_name in _extensions:
            # creation of a new extension, add entry in dict
            _extensions[extension_name] = dict()
        if _SETUP_FUNC_ATTR in _extensions[extension_name]:
            err = "There is already a setup function for this extension."
            raise ExtensionException(err)
        _extensions[extension_name][_SETUP_FUNC_ATTR] = func
        return func
    return register_setup_function


def extension_teardown(extension_name):
    """Returns a decorator to register a teardown function for an extension.

    :Parameters:
      extension_name
        String with the name of the extension.

    The decorated function will be called when the extension is deactivated.
    """
    def register_teardown_function(func):
        if not extension_name in _extensions:
            # creation of a new extension, add entry in dict
            _extensions[extension_name] = dict()
        if _TEARDOWN_FUNC_ATTR in _extensions[extension_name]:
            err = "There is already a teardown function for this extension."
            raise ExtensionException(err)
        _extensions[extension_name][_TEARDOWN_FUNC_ATTR] = func
        return func
    return register_teardown_function


class ExtensionNodeMetaclass(NodeMetaclass):
    """This is the metaclass for node extension superclasses.

    It takes care of registering extensions and the attributes in the
    extension.
    """

    def __new__(cls, classname, bases, members):
        """Create new node classes and register extensions.

        If a concrete extension node is created then a corresponding mixin
        class is automatically created and registered.
        """
        if classname == "ExtensionNode":
            # initial creation of ExtensionNode class
            return super(ExtensionNodeMetaclass, cls).__new__(cls, classname,
                                                              bases, members)
        # check if this is a new extension definition,
        # in that case this node is directly derived from ExtensionNode
        if ExtensionNode in bases:
            ext_name = members["extension_name"]
            if not ext_name:
                err = "No extension name has been specified."
                raise ExtensionException(err)
            if ext_name not in _extensions:
                # creation of a new extension, add entry in dict
                _extensions[ext_name] = dict()
            else:
                err = ("An extension with the name '" + ext_name +
                       "' has already been registered.")
                raise ExtensionException(err)
        # find the node that this extension node belongs to
        base_node_cls = None
        for base in bases:
            if type(base) is not ExtensionNodeMetaclass:
                if base_node_cls is None:
                    base_node_cls = base
                else:
                    err = ("Extension node derived from multiple "
                           "normal nodes.")
                    raise ExtensionException(err)
        if base_node_cls is None:
            # This new extension is not directly derived from another class,
            # so there is nothing to register (no default implementation).
            # We disable the doc method extension mechanism as this class
            # is not a node subclass and adding methods (e.g. _execute) would
            # cause problems.
            cls.DOC_METHODS = []
            return super(ExtensionNodeMetaclass, cls).__new__(cls, classname,
                                                              bases, members)
        ext_node_cls = super(ExtensionNodeMetaclass, cls).__new__(
                                                cls, classname, bases, members)
        ext_name = ext_node_cls.extension_name
        if not base_node_cls in _extensions[ext_name]:
            # register the base node
            _extensions[ext_name][base_node_cls] = dict()
        # Register methods from extension class hierarchy: iterate MRO in
        # reverse order and register all attributes starting from the
        # classes which are subclasses from ExtensionNode.
        extension_subtree = False
        for base in reversed(ext_node_cls.__mro__):
            # make sure we only inject methods in classes which have
            # ExtensionNode as superclass
            if extension_subtree and ExtensionNode in base.__mro__:
                for attr_name, attr_value in list(base.__dict__.items()):
                    if attr_name not in _NON_EXTENSION_ATTRIBUTES:
                        # check if this attribute has not already been
                        # extended in one of the base classes
                        already_active = False
                        for bb in ext_node_cls.__mro__:
                            if (bb in _extensions[ext_name] and
                            attr_name in _extensions[ext_name][bb] and
                            _extensions[ext_name][bb][attr_name] == attr_value):
                                already_active = True
                        # only register if not yet active
                        if not already_active:
                            _register_attribute(ext_name, base_node_cls,
                                                attr_name, attr_value)
            if base == ExtensionNode:
                extension_subtree = True
        return ext_node_cls


class ExtensionNode(with_metaclass(ExtensionNodeMetaclass, object)):
    """Base class for extensions nodes.

    A new extension node class should override the _extension_name.
    The concrete node implementations are then derived from this extension
    node class.

    To call an instance method from a parent class you have multiple options:

    - use super, but with the normal node class, e.g.:

      >>>  super(mdp.nodes.SFA2Node, self).method()      # doctest: +SKIP

      Here SFA2Node was given instead of the extension node class for the
      SFA2Node.

      If the extensions node class is used directly (without the extension
      mechanism) this can cause problems. In that case you have to be
      careful about the inheritance order and the effect on the MRO.

    - call it explicitly using the __func__ attribute [python version < 3]:

      >>> parent_class.method.__func__(self)             # doctest: +SKIP

      or [python version >=3]:

      >>> parent_class.method(self)                      # doctest: +SKIP

    To call the original (pre-extension) method in the same class use you
    simply prefix the method name with '_non_extension_' (this is the value
    of the `ORIGINAL_ATTR_PREFIX` constant in this module).
    """
    # override this name in a concrete extension node base class
    extension_name = None


def get_extensions():
    """Return a dictionary currently registered extensions.

    Note that this is not a copy, so if you change anything in this dict
    the whole extension mechanism will be affected. If you just want the
    names of the available extensions use get_extensions().keys().
    """
    return _extensions

def get_active_extensions():
    """Returns a list with the names of the currently activated extensions."""
    # use copy to protect the original set, also important if the return
    # value is used in a for-loop (see deactivate_extensions function)
    return list(_active_extensions)

def activate_extension(extension_name, verbose=False):
    """Activate the extension by injecting the extension methods."""
    if extension_name not in list(_extensions.keys()):
        err = "Unknown extension name: %s"%str(extension_name)
        raise ExtensionException(err)
    if extension_name in _active_extensions:
        if verbose:
            print('Extension %s is already active!' % extension_name)
        return
    _active_extensions.add(extension_name)
    try:
        if _SETUP_FUNC_ATTR in _extensions[extension_name]:
            _extensions[extension_name][_SETUP_FUNC_ATTR]()
        for node_cls, attributes in list(_extensions[extension_name].items()):
            if node_cls == _SETUP_FUNC_ATTR or node_cls == _TEARDOWN_FUNC_ATTR:
                continue
            for attr_name, attr_value in list(attributes.items()):
                if verbose:
                    print ("extension %s: adding %s to %s" %
                           (extension_name, attr_name, node_cls.__name__))
                ## store the original attribute / make it available
                ext_attr_name = _EXTENSION_ATTR_PREFIX + attr_name
                if attr_name in dir(node_cls):
                    if ext_attr_name in node_cls.__dict__:
                        # two extensions override the same attribute
                        err = ("Name collision for attribute '" +
                               attr_name + "' between extension '" +
                               getattr(node_cls, ext_attr_name)
                               + "' and newly activated extension '" +
                               extension_name + "'.")
                        raise ExtensionException(err)
                    # only overwrite the attribute if the extension is not
                    # yet active on this class or its superclasses
                    if ext_attr_name not in dir(node_cls):
                        original_attr = getattr(node_cls, attr_name)
                        if verbose:
                            print ("extension %s: overwriting %s in %s" %
                                (extension_name, attr_name, node_cls.__name__))
                        setattr(node_cls, ORIGINAL_ATTR_PREFIX + attr_name,
                                original_attr)
                setattr(node_cls, attr_name, attr_value)
                # store to which extension this attribute belongs, this is also
                # used as a flag that this is an extension attribute
                setattr(node_cls, ext_attr_name, extension_name)
    except Exception:
        # make sure that an incomplete activation is reverted
        deactivate_extension(extension_name)
        raise

def deactivate_extension(extension_name, verbose=False):
    """Deactivate the extension by removing the injected methods."""
    if extension_name not in list(_extensions.keys()):
        err = "Unknown extension name: " + str(extension_name)
        raise ExtensionException(err)
    if extension_name not in _active_extensions:
        return
    for node_cls, attributes in list(_extensions[extension_name].items()):
        if node_cls == _SETUP_FUNC_ATTR or node_cls == _TEARDOWN_FUNC_ATTR:
            continue
        for attr_name in list(attributes.keys()):
            original_name = ORIGINAL_ATTR_PREFIX + attr_name
            if verbose:
                print ("extension %s: removing %s from %s" %
                       (extension_name, attr_name, node_cls.__name__))
            if original_name in node_cls.__dict__:
                # restore the original attribute
                if verbose:
                    print ("extension %s: restoring %s in %s" %
                           (extension_name, attr_name, node_cls.__name__))
                delattr(node_cls, attr_name)
                original_attr = getattr(node_cls, original_name)
                # Check if the attribute is defined by one of the super
                # classes and test if the overwritten method is not that
                # method, otherwise we would inject unwanted methods.
                # Note: '==' tests identity for .__func__ and .__self__,
                #    but .im_class does not matter in Python 2.6.
                if all([getattr(x, attr_name, None) !=
                           original_attr for x in node_cls.__mro__[1:]]):
                    setattr(node_cls, attr_name, original_attr)
                delattr(node_cls, original_name)
            else:
                try:
                    # no original attribute to restore, so simply delete
                    # might be missing if the activation failed
                    delattr(node_cls, attr_name)
                except AttributeError:
                    pass
            try:
                # might be missing if the activation failed
                delattr(node_cls, _EXTENSION_ATTR_PREFIX + attr_name)
            except AttributeError:
                pass
    if _TEARDOWN_FUNC_ATTR in _extensions[extension_name]:
        _extensions[extension_name][_TEARDOWN_FUNC_ATTR]()
    _active_extensions.remove(extension_name)

def activate_extensions(extension_names, verbose=False):
    """Activate all the extensions for the given names.

    extension_names -- Sequence of extension names.
    """
    try:
        for extension_name in extension_names:
            activate_extension(extension_name, verbose=verbose)
    except:
        # if something goes wrong deactivate all, otherwise we might be
        # in an inconsistent state (e.g. methods for active extensions might
        # have been removed)
        deactivate_extensions(get_active_extensions())
        raise

def deactivate_extensions(extension_names, verbose=False):
    """Deactivate all the extensions for the given names.

    extension_names -- Sequence of extension names.
    """
    for extension_name in extension_names:
        deactivate_extension(extension_name, verbose=verbose)

# TODO: add check that only extensions are deactivated that were
#    originally activcated by this extension (same in context manager)
#    also add test for this
def with_extension(extension_name):
    """Return a wrapper function to activate and deactivate the extension.
    
    This function is intended to be used with the decorator syntax.

    The deactivation happens only if the extension was activated by
    the decorator (not if it was already active before). So this
    decorator ensures that the extensions is active and prevents
    unintended side effects.

    If the generated function is a generator, the extension will be in
    effect only when the generator object is created (that is when the
    function is called, but its body is not actually immediately
    executed). When the function body is executed (after ``next`` is
    called on the generator object), the extension might not be in
    effect anymore. Therefore, it is better to use the `extension`
    context manager with a generator function.
    """
    def decorator(func):
        def wrapper(*args, **kwargs):
            # make sure that we don't deactive and extension that was
            # not activated by the decorator (would be a strange sideeffect)
            if extension_name not in get_active_extensions():
                try:
                    activate_extension(extension_name)
                    result = func(*args, **kwargs)
                finally:
                    deactivate_extension(extension_name)
            else:
                result = func(*args, **kwargs)
            return result
        # now make sure that docstring and signature match the original
        func_info = NodeMetaclass._function_infodict(func)
        return NodeMetaclass._wrap_function(wrapper, func_info)
    return decorator

class extension(object):
    """Context manager for MDP extension.
    
    This allows you to use extensions using a ``with`` statement, as in:

    >>> with mdp.extension('extension_name'):
    ...     # 'node' is executed with the extension activated
    ...     node.execute(x)

    It is also possible to activate multiple extensions at once:

    >>> with mdp.extension(['ext1', 'ext2']):
    ...     # 'node' is executed with the two extensions activated
    ...     node.execute(x)
        
    The deactivation at the end happens only for the extensions that were
    activated by this context manager (not for those that were already active
    when the context was entered). This prevents unintended side effects.
    """

    def __init__(self, ext_names):
        if isinstance(ext_names, __builtins__['str']):
            ext_names = [ext_names]
        self.ext_names = ext_names
        self.deactivate_exts = []

    def __enter__(self):
        already_active = get_active_extensions()
        self.deactivate_exts = [ext_name for ext_name in self.ext_names
                                if ext_name not in already_active]
        activate_extensions(self.ext_names)

    def __exit__(self, type, value, traceback):
        deactivate_extensions(self.deactivate_exts)



from future import standard_library
standard_library.install_aliases()
from builtins import map
from builtins import str
from builtins import object
import types
import pickle
import mdp

class _Walk(object):
    """Recursively crawl an object and search for attributes that
    are reference to numpy arrays, return a dictionary:
    {attribute_name: array_reference}.

    Usage:
        _Walk()(object)
    """
    def __init__(self):
        self.arrays = {}
        self.start = None
        self.allobjs = {}

    def __call__(self, x, start = None):
        arrays = self.arrays
        # loop through the object dictionary
        for name in dir(x):
            # get the corresponding member
            obj = getattr(x, name)
            if id(obj) in list(self.allobjs.keys()):
                # if we already examined the member, skip to the next
                continue
            else:
                # add the id of this object to the list of know members
                self.allobjs[id(obj)] = None

            if start is None:
                # initialize a string structure to keep track of array names
                struct = name
            else:
                # struct is x.y.z (where x and y are objects and z an array)
                struct = '.'.join((start, name))

            if isinstance(obj, mdp.numx.ndarray):
                # the present member is an array
                # add it to the dictionary of all arrays
                if start is not None:
                    arrays[struct] = obj
                else:
                    arrays[name] = obj
            elif name.startswith('__') or type(obj) in (int, int, float,
                                                        types.MethodType):
                # the present member is a private member or a known
                # type that does not support arrays as attributes
                # Note: this is to avoid infinite
                # recursion in python2.6. Just remove the "or type in ..."
                # condition to see the error. There must be a better way.
                continue
            else:
                # we need to examine the present member in more detail
                arrays.update(self(obj, start = struct))
        self.start = start
        return arrays

def _format_dig(dict_):
    longest_name = max(list(map(len, list(dict_.keys()))))
    longest_size = max([len('%d'%x[0]) for x in list(dict_.values())])
    msgs = []
    total_size = 0
    for name in sorted(dict_.keys()):
        size = dict_[name][0]
        total_size += size
        pname = (name+':').ljust(longest_name+1)
        psize = ('%d bytes' % size).rjust(longest_size+6)
        msg = "%s %s" % (pname, psize)
        msgs.append(msg)
    final = "Total %d arrays (%d bytes)" % (len(dict_), total_size)
    msgs.append(final)
    return '\n'.join(msgs)

def dig_node(x):
    """Crawl recursively an MDP Node looking for arrays.

    Return (dictionary, string), where the dictionary is:
    { attribute_name: (size_in_bytes, array_reference)}
    and string is a nice string representation of it.
    """
    if not isinstance(x, mdp.Node):
        raise Exception('Cannot dig %s' % (str(type(x))))
    arrays = _Walk()(x)
    for name in list(arrays.keys()):
        ar = arrays[name]
        if len(ar.shape) == 0:
            size = 1
        else:
            size = mdp.numx.prod(ar.shape)
        bytes = ar.itemsize*size
        arrays[name] = (bytes, ar)
    return arrays, _format_dig(arrays)

def get_node_size(x):
    """Return node total byte-size using cPickle with protocol=2.

    The byte-size is related to the memory needed by the node).
    """
    # TODO: add check for problematic node types, like NoiseNode?
    # TODO: replace this with sys.getsizeof for Python >= 2.6
    size = len(pickle.dumps(x, protocol = 2))
    return size

def get_node_size_str(x, si_units=False):
    """Return node total byte-size as a well readable string.

    si_units -- If True si-units like KB are used instead of KiB.

    The get_node_size function is used to get the size.
    """
    return _memory_size_str(get_node_size(x), si_units=si_units)

_SI_MEMORY_PREFIXES = ("", "k", "M", "G", "T", "P", "E")
_IEC_MEMORY_PREFIXES = ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei")

def _memory_size_str(size, si_units=False):
    """Convert the given memory size into a nicely formatted string.

    si_units -- If True si-units like kB are used instead of kiB.
    """
    if si_units:
        base = 10**3
    else:
        base = 2**10
    scale = 0  # 1024**scale is the actual scale
    while size > base**(scale+1):
        scale += 1
    unit = "B"
    if scale:
        size_str = size = "%.1f" % (1.0 * size / (base**scale))
        if si_units:
            unit = _SI_MEMORY_PREFIXES[scale] + unit
        else:
            unit = _IEC_MEMORY_PREFIXES[scale] + unit
    else:
        size_str = "%d" % size
    return size_str + " " + unit



from __future__ import division
from builtins import map
from builtins import range
from past.utils import old_div
from builtins import object
__docformat__ = "restructuredtext en"

from mdp import numx, numx_rand, utils, graph, Node

class _NGNodeData(object):
    """Data associated to a node in a Growing Neural Gas graph."""
    def __init__(self, pos, error=0.0, hits=0, label=None):
        # reference vector (spatial position)
        self.pos = pos
        # cumulative error
        self.cum_error = error
        self.hits = hits
        self.label = label

class _NGEdgeData(object):
    """Data associated to an edge in a Growing Neural Gas graph."""
    def __init__(self, age=0):
        self.age = age
    def inc_age(self):
        self.age += 1


class GrowingNeuralGasNode(Node):
    """Learn the topological structure of the input data by building a
    corresponding graph approximation.

    The algorithm expands on the original Neural Gas algorithm
    (see mdp.nodes NeuralGasNode) in that the algorithm adds new nodes are
    added to the graph as more data becomes available. Im this way,
    if the growth rate is appropriate, one can avoid overfitting  or
    underfitting the data.

    More information about the Growing Neural Gas algorithm can be found in
    B. Fritzke, A Growing Neural Gas Network Learns Topologies, in G. Tesauro,
    D. S. Touretzky, and T. K. Leen (editors), Advances in Neural Information
    Processing Systems 7, pages 625-632. MIT Press, Cambridge MA, 1995.

    **Attributes and methods of interest**

    - graph -- The corresponding `mdp.graph.Graph` object
    """
    def __init__(self, start_poss=None, eps_b=0.2, eps_n=0.006, max_age=50,
                 lambda_=100, alpha=0.5, d=0.995, max_nodes=2147483647,
                 input_dim=None, dtype=None):
        """Growing Neural Gas algorithm.

        :Parameters:

          start_poss
            sequence of two arrays containing the position of the
            first two nodes in the GNG graph. If unspecified, the
            initial nodes are chosen with a random position generated
            from a gaussian distribution with zero mean and unit
            variance.

          eps_b
            coefficient of movement of the nearest node to a new data
            point. Typical values are 0 < eps_b << 1 .

            Default: 0.2

          eps_n
            coefficient of movement of the neighbours of the nearest
            node to a new data point. Typical values are
            0 < eps_n << eps_b .

            Default: 0.006

          max_age
            remove an edge after `max_age` updates. Typical values are
            10 < max_age < lambda.

            Default: 50

          `lambda_`
            insert a new node after `lambda_` steps. Typical values are O(100).

            Default: 100

          alpha
            when a new node is inserted, multiply the error of the
            nodes from which it generated by 0<alpha<1. A typical value
            is 0.5.

            Default: 0.5

          d
            each step the error of the nodes are multiplied by 0<d<1.
            Typical values are close to 1.

            Default: 0.995

          max_nodes
            maximal number of nodes in the graph.

            Default: 2^31 - 1
        """
        self.graph = graph.Graph()
        self.tlen = 0

        #copy parameters
        (self.eps_b, self.eps_n, self.max_age, self.lambda_, self.alpha,
         self.d, self.max_nodes) = (eps_b, eps_n, max_age, lambda_, alpha,
                                    d, max_nodes)

        super(GrowingNeuralGasNode, self).__init__(input_dim, None, dtype)

        if start_poss is not None:
            if self.dtype is None:
                self.dtype = start_poss[0].dtype
            node1 = self._add_node(self._refcast(start_poss[0]))
            node2 = self._add_node(self._refcast(start_poss[1]))
            self._add_edge(node1, node2)

    def _set_input_dim(self, n):
        self._input_dim = n
        self.output_dim = n

    def _add_node(self, pos):
        node = self.graph.add_node(_NGNodeData(pos))
        return node

    def _add_edge(self, from_, to_):
        self.graph.add_edge(from_, to_, _NGEdgeData())

    def _get_nearest_nodes(self, x):
        """Return the two nodes in the graph that are nearest to x and their
        squared distances. (Return ([node1, node2], [dist1, dist2])"""
        # distance function
        def _distance_from_node(node):
            #return norm(node.data.pos-x)**2
            tmp = node.data.pos - x
            return utils.mult(tmp, tmp)
        g = self.graph
        # distances of all graph nodes from x
        distances = numx.array(list(map(_distance_from_node, g.nodes)))
        ids = distances.argsort()[:2]
        #nearest = [g.nodes[idx] for idx in ids]
        #return nearest, distances[ids]
        return (g.nodes[ids[0]], g.nodes[ids[1]]), distances.take(ids)

    def _move_node(self, node, x, eps):
        """Move a node by eps in the direction x."""
        # ! make sure that eps already has the right dtype
        node.data.pos += eps*(x - node.data.pos)

    def _remove_old_edges(self, edges):
        """Remove all edges older than the maximal age."""
        g, max_age = self.graph, self.max_age
        for edge in edges:
            if edge.data.age > max_age:
                g.remove_edge(edge)
                if edge.head.degree() == 0:
                    g.remove_node(edge.head)
                if edge.tail.degree() == 0:
                    g.remove_node(edge.tail)

    def _insert_new_node(self):
        """Insert a new node in the graph where it is more necessary (i.e.
        where the error is the largest)."""
        g = self.graph
        # determine the node with the highest error
        errors = [x.data.cum_error for x in g.nodes]
        qnode = g.nodes[numx.argmax(errors)]
        # determine the neighbour with the highest error
        neighbors = qnode.neighbors()
        errors = [x.data.cum_error for x in neighbors]
        fnode = neighbors[numx.argmax(errors)]
        # new node, halfway between the worst node and the worst of
        # its neighbors
        new_pos = 0.5*(qnode.data.pos + fnode.data.pos)
        new_node = self._add_node(new_pos)
        # update edges
        edges = qnode.get_edges(neighbor=fnode)
        g.remove_edge(edges[0])
        self._add_edge(qnode, new_node)
        self._add_edge(fnode, new_node)
        # update errors
        qnode.data.cum_error *= self.alpha
        fnode.data.cum_error *= self.alpha
        new_node.data.cum_error = 0.5*(qnode.data.cum_error+
                                       fnode.data.cum_error)

    def get_nodes_position(self):
        return numx.array([n.data.pos for n in self.graph.nodes],
                          dtype = self.dtype)

    def _train(self, input):
        g = self.graph
        d = self.d

        if len(g.nodes)==0:
            # if missing, generate two initial nodes at random
            # assuming that the input data has zero mean and unit variance,
            # choose the random position according to a gaussian distribution
            # with zero mean and unit variance
            normal = numx_rand.normal
            self._add_node(self._refcast(normal(0.0, 1.0, self.input_dim)))
            self._add_node(self._refcast(normal(0.0, 1.0, self.input_dim)))

        # loop on single data points
        for x in input:
            self.tlen += 1

            # step 2 - find the nearest nodes
            # dists are the squared distances of x from n0, n1
            (n0, n1), dists = self._get_nearest_nodes(x)

            # step 3 - increase age of the emanating edges
            for e in n0.get_edges():
                e.data.inc_age()

            # step 4 - update error
            n0.data.cum_error += numx.sqrt(dists[0])

            # step 5 - move nearest node and neighbours
            self._move_node(n0, x, self.eps_b)
            # neighbors undirected
            neighbors = n0.neighbors()
            for n in neighbors:
                self._move_node(n, x, self.eps_n)

            # step 6 - update n0<->n1 edge
            if n1 in neighbors:
                # should be one edge only
                edges = n0.get_edges(neighbor=n1)
                edges[0].data.age = 0
            else:
                self._add_edge(n0, n1)

            # step 7 - remove old edges
            self._remove_old_edges(n0.get_edges())

            # step 8 - add a new node each lambda steps
            if not self.tlen % self.lambda_ and len(g.nodes) < self.max_nodes:
                self._insert_new_node()

            # step 9 - decrease errors
            for node in g.nodes:
                node.data.cum_error *= d

    def nearest_neighbor(self, input):
        """Assign each point in the input data to the nearest node in
        the graph. Return the list of the nearest node instances, and
        the list of distances.
        Executing this function will close the training phase if
        necessary."""
        super(GrowingNeuralGasNode, self).execute(input)

        nodes = []
        dists = []
        for x in input:
            (n0, _), dist = self._get_nearest_nodes(x)
            nodes.append(n0)
            dists.append(numx.sqrt(dist[0]))
        return nodes, dists

class NeuralGasNode(GrowingNeuralGasNode):
    """Learn the topological structure of the input data by building a
    corresponding graph approximation (original Neural Gas algorithm).

    The Neural Gas algorithm was originally published in Martinetz, T. and
    Schulten, K.: A "Neural-Gas" Network Learns Topologies. In Kohonen, T.,
    Maekisara, K., Simula, O., and Kangas, J. (eds.), Artificial Neural
    Networks. Elsevier, North-Holland., 1991.

    **Attributes and methods of interest**

    - graph -- The corresponding `mdp.graph.Graph` object
    - max_epochs - maximum number of epochs until which to train.
    """
    def __init__(self, num_nodes = 10,
                       start_poss=None,
                       epsilon_i=0.3,               # initial epsilon
                       epsilon_f=0.05,              # final epsilon
                       lambda_i=30.,                # initial lambda
                       lambda_f=0.01,               # final lambda
                       max_age_i=20,                # initial edge lifetime
                       max_age_f=200,               # final edge lifetime
                       max_epochs=100,
                       n_epochs_to_train=None,
                       input_dim=None,
                       dtype=None):
        """Neural Gas algorithm.

        Default parameters taken from the original publication.

        :Parameters:

          start_poss
            sequence of two arrays containing the position of the
            first two nodes in the GNG graph. In unspecified, the
            initial nodes are chosen with a random position generated
            from a gaussian distribution with zero mean and unit
            variance.

          num_nodes
            number of nodes to use. Ignored if start_poss is given.

          epsilon_i, epsilon_f
            initial and final values of epsilon. Fraction of the distance
            between the closest node and the presented data point by which the
            node moves towards the data point in an adaptation step. Epsilon
            decays during training by e(t) = e_i(e_f/e_i)^(t/t_max) with t
            being the epoch.

          lambda_i, lambda_f
            initial and final values of lambda. Lambda influences how the
            weight change of nodes in the ranking decreases with lower rank. It 
            is sometimes called the "neighborhood factor". Lambda decays during
            training in the same manner as epsilon does.

          max_age_i, max_age_f
            Initial and final lifetime, after which an edge will be removed.
            Lifetime is measured in terms of adaptation steps, i.e.,
            presentations of data points. It decays during training like
            epsilon does.

          max_epochs
            number of epochs to train. One epoch has passed when all data points
            from the input have been presented once. The default in the original 
            publication was 40000, but since this has proven to be impractically
            high too high for many real-world data sets, we adopted a default
            value of 100.

          n_epochs_to_train
            number of epochs to train on each call. Useful for batch learning
            and for visualization of the training process. Default is to
            train once until max_epochs is reached.
        """

        self.graph = graph.Graph()

        if n_epochs_to_train is None:
            n_epochs_to_train = max_epochs

        #copy parameters
        self.num_nodes = num_nodes
        self.start_poss = start_poss
        self.epsilon_i = epsilon_i
        self.epsilon_f = epsilon_f
        self.lambda_i = lambda_i
        self.lambda_f = lambda_f
        self.max_age_i = max_age_i
        self.max_age_f = max_age_f
        self.max_epochs = max_epochs
        self.n_epochs_to_train = n_epochs_to_train

        super(GrowingNeuralGasNode, self).__init__(input_dim, None, dtype)

        if start_poss is not None:
            if self.num_nodes != len(start_poss):
                self.num_nodes = len(start_poss)
            if self.dtype is None:
                self.dtype = start_poss[0].dtype
            for node_ind in range(self.num_nodes):
                self._add_node(self._refcast(start_poss[node_ind]))
        self.epoch = 0


    def _train(self, input):
        g = self.graph

        if len(g.nodes) == 0:
            # if missing, generate num_nodes initial nodes at random
            # assuming that the input data has zero mean and unit variance,
            # choose the random position according to a gaussian distribution
            # with zero mean and unit variance
            normal = numx_rand.normal
            for _ in range(self.num_nodes):
                self._add_node(self._refcast(normal(0.0, 1.0, self.input_dim)))

        epoch = self.epoch
        e_i = self.epsilon_i
        e_f = self.epsilon_f
        l_i = self.lambda_i
        l_f = self.lambda_f
        T_i = float(self.max_age_i)
        T_f = float(self.max_age_f)
        max_epochs = float(self.max_epochs)
        remaining_epochs = self.n_epochs_to_train
        while remaining_epochs > 0:
            # reset permutation of data points
            di = numx.random.permutation(input)
            if epoch < max_epochs:
                denom = old_div(epoch,max_epochs)
            else:
                denom = 1.
            epsilon = e_i * ((old_div(e_f,e_i))**denom)
            lmbda = l_i * ((old_div(l_f,l_i))**denom)
            T = T_i * ((old_div(T_f,T_i))**denom)
            epoch += 1
            for x in di:
                # Step 1 rank nodes according to their distance to random point
                ranked_nodes = self._rank_nodes_by_distance(x)

                # Step 2 move nodes
                for rank,node in enumerate(ranked_nodes):
                    #TODO: cut off at some rank when using many nodes
                    #TODO: check speedup by vectorizing
                    delta_w = epsilon * numx.exp(old_div(-rank, lmbda)) * \
                                    (x - node.data.pos)
                    node.data.pos += delta_w

                # Step 3 update edge weight
                for e in g.edges:
                    e.data.inc_age()

                # Step 4 set age of edge between first two nodes to zero
                #  or create it if it doesn't exist.
                n0 = ranked_nodes[0]
                n1 = ranked_nodes[1]
                nn = n0.neighbors()
                if n1 in nn:
                    edges = n0.get_edges(neighbor=n1)
                    edges[0].data.age = 0  # should only be one edge
                else:
                    self._add_edge(n0, n1)

                # step 5 delete edges with age > max_age
                self._remove_old_edges(max_age=T)
            remaining_epochs -= 1
        self.epoch = epoch


    def _rank_nodes_by_distance(self, x):
        """Return the nodes in the graph in a list ranked by their squared
        distance to x. """

        #TODO: Refactor together with GNGNode._get_nearest_nodes

        # distance function
        def _distance_from_node(node):
            tmp = node.data.pos - x
            return utils.mult(tmp, tmp) # maps to mdp.numx.dot

        g = self.graph

        # distances of all graph nodes from x
        distances = numx.array(list(map(_distance_from_node, g.nodes)))
        ids = distances.argsort()
        ranked_nodes = [g.nodes[id] for id in ids]

        return ranked_nodes


    def _remove_old_edges(self, max_age):
        """Remove edges with age > max_age."""
        g = self.graph
        for edge in self.graph.edges:
            if edge.data.age > max_age:
                g.remove_edge(edge)



from ._tools import *

def testHistogramNode_nofraction():
    """Test HistogramNode with fraction set to 1.0."""
    node = mdp.nodes.HistogramNode()
    x1 = numx.array([[0.1, 0.2], [0.3, 0.5]])
    x2 = numx.array([[0.3, 0.6], [0.2, 0.1]])
    x = numx.concatenate([x1, x2])
    node.train(x1)
    node.train(x2)
    assert numx.all(x == node.data_hist)

def testHistogramNode_fraction():
    """Test HistogramNode with fraction set to 0.5."""
    node = mdp.nodes.HistogramNode(hist_fraction=0.5)
    x1 = numx_rand.random((1000, 3))
    x2 = numx_rand.random((500, 3))
    node.train(x1)
    node.train(x2)
    assert len(node.data_hist) < 1000



from __future__ import division
from builtins import map
from builtins import next
from builtins import range
from past.utils import old_div
import mdp

# import numeric module (scipy, Numeric or numarray)
numx, numx_rand, numx_linalg  = mdp.numx, mdp.numx_rand, mdp.numx_linalg
numx_description = mdp.numx_description
import random
import itertools

def timediff(data):
    """Returns the array of the time differences of data."""
    # this is the fastest way we found so far
    return data[1:]-data[:-1]

def refcast(array, dtype):
    """
    Cast the array to dtype only if necessary, otherwise return a reference.
    """
    dtype = mdp.numx.dtype(dtype)
    if array.dtype == dtype:
        return array
    return array.astype(dtype)

def scast(scalar, dtype):
    """Convert a scalar in a 0D array of the given dtype."""
    return numx.array(scalar, dtype=dtype)

def rotate(mat, angle, columns=(0, 1), units='radians'):
    """
    Rotate in-place data matrix (NxM) in the plane defined by the columns=[i,j]
    when observation are stored on rows. Observations are rotated
    counterclockwise. This corresponds to the following matrix-multiplication
    for each data-point (unchanged elements omitted):

     [  cos(angle) -sin(angle)     [ x_i ]
        sin(angle)  cos(angle) ] * [ x_j ]

    If M=2, columns=[0,1].
    """
    if units is 'degrees':
        angle = angle/180.*numx.pi
    cos_ = numx.cos(angle)
    sin_ = numx.sin(angle)
    [i, j] = columns
    col_i = mat[:, i] + 0.
    col_j = mat[:, j]
    mat[:, i] = cos_*col_i - sin_*col_j
    mat[:, j] = sin_*col_i + cos_*col_j

def permute(x, indices=(0, 0), rows=0, cols=1):
    """Swap two columns and (or) two rows of 'x', whose indices are specified
    in indices=[i,j].
    Note: permutations are done in-place. You'll lose your original matrix"""
    ## the nicer option:
    ## x[i,:],x[j,:] = x[j,:],x[i,:]
    ## does not work because array-slices are references.
    ## The following would work:
    ## x[i,:],x[j,:] = x[j,:].tolist(),x[i,:].tolist()
    ## because list-slices are copies, but you get 2
    ## copies instead of the one you need with our method.
    ## This would also work:
    ## tmp = x[i,:].copy()
    ## x[i,:],x[j,:] = x[j,:],tmp
    ## but it is slower (for larger matrices) than the one we use.
    [i, j] = indices
    if rows:
        x[i, :], x[j, :] = x[j, :], x[i, :] + 0
    if cols:
        x[:, i], x[:, j] = x[:, j], x[:, i] + 0

def hermitian(x):
    """Compute the Hermitian, i.e. conjugate transpose, of x."""
    return x.T.conj()

def symrand(dim_or_eigv, dtype="d"):
    """Return a random symmetric (Hermitian) matrix.

    If 'dim_or_eigv' is an integer N, return a NxN matrix, with eigenvalues
        uniformly distributed on (-1,1).

    If 'dim_or_eigv' is  1-D real array 'a', return a matrix whose
                      eigenvalues are 'a'.
    """
    if isinstance(dim_or_eigv, int):
        dim = dim_or_eigv
        d = (numx_rand.random(dim)*2) - 1
    elif isinstance(dim_or_eigv,
                    numx.ndarray) and len(dim_or_eigv.shape) == 1:
        dim = dim_or_eigv.shape[0]
        d = dim_or_eigv
    else:
        raise mdp.MDPException("input type not supported.")

    v = random_rot(dim)
    #h = mdp.utils.mult(mdp.utils.mult(hermitian(v), mdp.numx.diag(d)), v)
    h = mdp.utils.mult(mult_diag(d, hermitian(v), left=False), v)
    # to avoid roundoff errors, symmetrize the matrix (again)
    h = 0.5*(h.T+h)
    if dtype in ('D', 'F', 'G'):
        h2 = symrand(dim_or_eigv)
        h = h + 1j*(numx.triu(h2)-numx.tril(h2))
    return refcast(h, dtype)

def random_rot(dim, dtype='d'):
    """Return a random rotation matrix, drawn from the Haar distribution
    (the only uniform distribution on SO(n)).
    The algorithm is described in the paper
    Stewart, G.W., "The efficient generation of random orthogonal
    matrices with an application to condition estimators", SIAM Journal
    on Numerical Analysis, 17(3), pp. 403-409, 1980.
    For more information see
    http://en.wikipedia.org/wiki/Orthogonal_matrix#Randomization"""
    H = mdp.numx.eye(dim, dtype=dtype)
    D = mdp.numx.ones((dim,), dtype=dtype)
    for n in range(1, dim):
        x = mdp.numx_rand.normal(size=(dim-n+1,)).astype(dtype)
        D[n-1] = mdp.numx.sign(x[0])
        x[0] -= D[n-1]*mdp.numx.sqrt((x*x).sum())
        # Householder transformation
        Hx = ( mdp.numx.eye(dim-n+1, dtype=dtype)
               - 2.*mdp.numx.outer(x, x)/(x*x).sum() )
        mat = mdp.numx.eye(dim, dtype=dtype)
        mat[n-1:, n-1:] = Hx
        H = mdp.utils.mult(H, mat)
    # Fix the last sign such that the determinant is 1
    D[-1] = (-1)**(1-dim%2)*D.prod()
    # Equivalent to mult(numx.diag(D), H) but faster
    H = (D*H.T).T
    return H

def norm2(v):
    """Compute the 2-norm for 1D arrays.
    norm2(v) = sqrt(sum(v_i^2))"""

    return numx.sqrt((v*v).sum())

def cov2(x, y):
    """Compute the covariance between 2D matrices x and y.
    Complies with the old scipy.cov function: different variables
    are on different columns."""

    mnx = x.mean(axis=0)
    mny = y.mean(axis=0)
    tlen = x.shape[0]
    return old_div(mdp.utils.mult(x.T, y),(tlen-1)) - numx.outer(mnx, mny)

def cov_maxima(cov):
    """Extract the maxima of a covariance matrix."""
    dim = cov.shape[0]
    maxs = []
    if dim >= 1:
        cov=abs(cov)
        glob_max_idx = (cov.argmax()//dim, cov.argmax()%dim)
        maxs.append(cov[glob_max_idx[0], glob_max_idx[1]])
        cov_reduce = cov.copy()
        cov_reduce = cov_reduce[numx.arange(dim) != glob_max_idx[0], :]
        cov_reduce = cov_reduce[:, numx.arange(dim) != glob_max_idx[1]]
        maxs.extend(cov_maxima(cov_reduce))
        return maxs
    else:
        return []


def mult_diag(d, mtx, left=True):
    """Multiply a full matrix by a diagonal matrix.
    This function should always be faster than dot.

    Input:
      d -- 1D (N,) array (contains the diagonal elements)
      mtx -- 2D (N,N) array

    Output:
      mult_diag(d, mts, left=True) == dot(diag(d), mtx)
      mult_diag(d, mts, left=False) == dot(mtx, diag(d))
    """
    if left:
        return (d*mtx.T).T
    else:
        return d*mtx

def comb(N, k):
    """Return number of combinations of k objects from a set of N objects
    without repetitions, a.k.a. the binomial coefficient of N and k."""
    ret = 1
    for mlt in range(N, N-k, -1):
        ret *= mlt
    for dv in range(1, k+1):
        ret //= dv
    return ret

# WARNING numpy.linalg.eigh does not support float sizes larger than 64 bits,
# and complex numbers of size larger than 128 bits.
# Also float16 is not supported either.
# This is not a problem for MDP, as long as scipy.linalg.eigh is available.
def get_dtypes(typecodes_key, _safe=True):
    """Return the list of dtypes corresponding to the set of
    typecodes defined in numpy.typecodes[typecodes_key].
    E.g., get_dtypes('Float') = [dtype('f'), dtype('d'), dtype('g')].

    If _safe is True (default), we remove large floating point types
    if the numerical backend does not support them.
    """
    types = []
    for c in numx.typecodes[typecodes_key]:
        try:
            type_ = numx.dtype(c)
            if (_safe and not mdp.config.has_symeig == 'scipy.linalg.eigh'
                and type_ in _UNSAFE_DTYPES):
                continue
            types.append(type_)
        except TypeError:
            pass
    return types

_UNSAFE_DTYPES = [numx.typeDict[d] for d in
                  ['float16', 'float96', 'float128', 'complex192', 'complex256']
                  if d in numx.typeDict]

def nongeneral_svd(A, range=None, **kwargs):
    """SVD routine for simple eigenvalue problem, API is compatible with
    symeig."""
    Z2, w, Z = mdp.utils.svd(A)
    # sort eigenvalues and corresponding eigenvectors
    idx = w.argsort()
    w = w.take(idx)
    Z = Z.take(idx, axis=0).T
    if range is not None:
        lo, hi = range
        Z = Z[:, lo-1:hi]
        w = w[lo-1:hi]
    return w, Z

def sqrtm(A):
    """This is a symmetric definite positive matrix sqrt function"""
    d, V = mdp.utils.symeig(A)
    return mdp.utils.mult(V, mult_diag(numx.sqrt(d), V.T))

# replication functions
def lrep(x, n):
    """Replicate x n-times on a new first dimension"""
    shp = [1]
    shp.extend(x.shape)
    return x.reshape(shp).repeat(n, axis=0)

def rrep(x, n):
    """Replicate x n-times on a new last dimension"""
    shp = x.shape + (1,)
    return x.reshape(shp).repeat(n, axis=-1)

def irep(x, n, dim):
    """Replicate x n-times on a new dimension dim-th dimension"""
    x_shape = x.shape
    shp = x_shape[:dim] + (1,) + x_shape[dim:]
    return x.reshape(shp).repeat(n, axis=dim)
# /replication functions

try:
    # product exists only in itertools >= 2.6
    from itertools import product
except ImportError:
    def product(*args, **kwds):
        """Cartesian product of input iterables.
        """
        # taken from python docs 2.6
        # product('ABCD', 'xy') --> Ax Ay Bx By Cx Cy Dx Dy
        # product(range(2), repeat=3) --> 000 001 010 011 100 101 110 111
        pools = list(map(tuple, args)) * kwds.get('repeat', 1)
        result = [[]]
        for pool in pools:
            result = [x+[y] for x in result for y in pool]
        for prod in result:
            yield tuple(prod)

def orthogonal_permutations(a_dict):
    """
    Takes a dictionary with lists as keys and returns all permutations
    of these list elements in new dicts.

    This function is useful, when a method with several arguments
    shall be tested and all of the arguments can take several values.

    The order is not defined, therefore the elements should be
    orthogonal to each other.

    >>> for i in orthogonal_permutations({'a': [1,2,3], 'b': [4,5]}):
            print i
    {'a': 1, 'b': 4}
    {'a': 1, 'b': 5}
    {'a': 2, 'b': 4}
    {'a': 2, 'b': 5}
    {'a': 3, 'b': 4}
    {'a': 3, 'b': 5}
    """
    pool = dict(a_dict)
    args = []
    for func, all_args in list(pool.items()):
        # check the size of the list in the second item of the tuple
        args_with_fun = [(func, arg) for arg in all_args]
        args.append(args_with_fun)
    for i in product(*args):
        yield dict(i)


def izip_stretched(*iterables):
    """Same as izip, except that for convenience non-iterables are repeated ad infinitum.

    This is useful when trying to zip input data with respective labels
    and allows for having a single label for all data, as well as for
    havning a list of labels for each data vector.
    Note that this will take strings as an iterable (of course), so
    strings acting as a single value need to be wrapped in a repeat
    statement of their own.

    Thus,
    >>> for zipped in izip_stretched([1, 2, 3], -1):
            print zipped
    (1, -1)
    (2, -1)
    (3, -1)

    is equivalent to
    >>> for zipped in izip([1, 2, 3], [-1] * 3):
            print zipped
    (1, -1)
    (2, -1)
    (3, -1)
    """
    def iter_or_repeat(val):
        try:
            return iter(val)
        except TypeError:
            return itertools.repeat(val)

    iterables= list(map(iter_or_repeat, iterables))
    while iterables:
        # need to care about python < 2.6
        yield tuple([next(it) for it in iterables])


def weighted_choice(a_dict, normalize=True):
    """Returns a key from a dictionary based on the weight that the value suggests.
    If 'normalize' is False, it is assumed the weights sum up to unity. Otherwise,
    the algorithm will take care of normalising.

    Example:
    >>> d = {'a': 0.1, 'b': 0.5, 'c': 0.4}
    >>> weighted_choice(d)
    # draws 'b':'c':'a' with 5:4:1 probability

    TODO: It might be good to either shuffle the order or explicitely specify it,
    before walking through the items, to minimise possible degeneration.
    """
    if normalize:
        d = a_dict.copy()
        s = sum(d.values())
        for key, val in list(d.items()):
            d[key] = old_div(d[key], s)
    else:
        d = a_dict
    rand_num = random.random()
    total_rand = 0
    for key, val in list(d.items()):
        total_rand += val
        if total_rand > rand_num:
            return key
    return None

def bool_to_sign(an_array):
    """Return -1 for each False; +1 for each True"""
    return numx.sign(an_array - 0.5)

def sign_to_bool(an_array, zero=True):
    """Return False for each negative value, else True.

    The value for 0 is specified with 'zero'.
    """
    if zero:
        return numx.array(an_array) >= 0
    else:
        return numx.array(an_array) > 0

def gabor(size, alpha, phi, freq, sgm, x0=None, res=1, ampl=1.):
    """Return a 2D array containing a Gabor wavelet.

    Input arguments:
    size -- (height, width) (pixels)
    alpha -- orientation (rad)
    phi -- phase (rad)
    freq -- frequency (cycles/deg)
    sgm -- (sigma_x, sigma_y) standard deviation along the axis
           of the gaussian ellipse (pixel)
    x0 -- (x,y) coordinates of the center of the wavelet (pixel)
          Default: None, meaning the center of the array
    res -- spatial resolution (deg/pixel)
           Default: 1, so that 'freq' is measured in cycles/pixel
    ampl -- constant multiplying the result
            Default: 1.
    """

    # init
    w, h = size
    if x0 is None: x0 = (w//2, h//2)
    y0, x0 = x0

    # some useful quantities
    freq *= res
    sinalpha = numx.sin(alpha)
    cosalpha = numx.cos(alpha)
    v0, u0 = freq*cosalpha, freq*sinalpha

    # coordinates
    #x = numx.mgrid[-x0:w-x0, -y0:h-y0]
    x = numx.meshgrid(numx.arange(w)-x0, numx.arange(h)-y0)
    x = (x[0].T, x[1].T)
    xr = x[0]*cosalpha - x[1]*sinalpha
    yr = x[0]*sinalpha + x[1]*cosalpha

    # gabor
    im = ampl*numx.exp(-0.5*(xr*xr/(sgm[0]*sgm[0]) + yr*yr/(sgm[1]*sgm[1]))) \
             *numx.cos(-2.*numx.pi*(u0*x[0]+v0*x[1]) - phi)

    return im

def residuals(app_x, y_noisy, exp_funcs, x_orig, k=0.0):
    """Function used internally by invert_exp_funcs2 to approximate
    inverses in GeneralExpansionNode. """
    app_x = app_x.reshape((1,-1))
    app_exp_x =  numx.concatenate([func(app_x) for func in exp_funcs],axis=1)

    div_y = numx.sqrt(len(y_noisy))
    div_x = numx.sqrt(len(x_orig))  
    return numx.append( (1-k)**0.5 *(y_noisy-app_exp_x[0]) / div_y, (k)**0.5 * (x_orig - app_x[0])/div_x )

def invert_exp_funcs2(exp_x_noisy, dim_x, exp_funcs, use_hint=False, k=0.0):
    """Approximates a preimage app_x of exp_x_noisy.

    Returns an array app_x, such that each row of exp_funcs(app_x) is close
    to each row of exp_x_noisy.

    use_hint: determines the starting point for the approximation of the
    preimage. There are three possibilities.
    if it equals False: starting point is generated with a normal distribution
    if it equals True: starting point is the first dim_x elements of exp_x_noisy
    otherwise: use the parameter use_hint itself as the first approximation

    k: weighting factor in [0, 1] to balance between approximation error and
       closeness to the starting point. For instance:
       objective function is to minimize:
           (1-k) * ||exp_funcs(app_x) - exp_x_noisy||**2/output_dim +
               k * ||app_x - starting point||**2/input_dim

    Note: this function requires scipy.
    """
    if numx_description != 'scipy':
        raise NotImplementedError('This function requires scipy.')
    else:
        import scipy.optimize
    num_samples = exp_x_noisy.shape[0]

    if isinstance(use_hint, numx.ndarray):
        app_x = use_hint.copy()
    elif use_hint == True:
        app_x = exp_x_noisy[:,0:dim_x].copy()
    else:
        app_x = numx.random.normal(size=(num_samples,dim_x))

    for row in range(num_samples):
        plsq = scipy.optimize.leastsq(residuals, app_x[row],
                                      args=(exp_x_noisy[row], exp_funcs,
                                            app_x[row], k), maxfev=50*dim_x)
        app_x[row] = plsq[0]

    app_exp_x = numx.concatenate([func(app_x) for func in exp_funcs],axis=1)
    return app_x, app_exp_x



from .caching_extension import (activate_caching, deactivate_caching,
                               cache, set_cachedir,
                               __doc__, __docformat__)

from mdp.utils import fixup_namespace

__all__ = ['activate_caching', 'deactivate_caching',
           'cache', 'set_cachedir']

fixup_namespace(__name__, __all__,('caching_extension','fixup_namespace',))



"""
Module with simple functions for the complete inspection procedure.
"""

from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import str

import os
import webbrowser
import pickle as pickle
import tempfile
import traceback
import warnings

import mdp
from mdp import numx

from bimdp import BiFlow

from .tracer import (
    InspectionHTMLTracer, TraceDebugException, inspection_css,
    prepare_training_inspection, remove_inspection_residues,
    _trace_biflow_training, PICKLE_EXT, STANDARD_CSS_FILENAME
)
from .slideshow import (
    TrainHTMLSlideShow, ExecuteHTMLSlideShow,
    SectExecuteHTMLSlideShow)
from .utils import robust_write_file, robust_pickle, first_iterable_elem


def _open_custom_brower(open_browser, url):
    """Helper function to support opening a custom browser."""
    if isinstance(open_browser, str):
        try:
            custom_browser = webbrowser.get(open_browser)
            custom_browser.open(url)
        except webbrowser.Error:
            err = ("Could not open browser '%s', using default." %
                   open_browser)
            warnings.warn(err)
            webbrowser.open(url)
    else:
        webbrowser.open(url)

def standard_css():
    """Return the standard CSS for inspection."""
    return (mdp.utils.basic_css() + inspection_css() +
            TrainHTMLSlideShow.slideshow_css())


class EmptyTraceException(Exception):
    """Exception for empty traces, i.e., when no slides where generated."""
    pass


def inspect_training(snapshot_path, x_samples, msg_samples=None,
                     stop_messages=None, inspection_path=None,
                     tracer=None, debug=False,
                     slide_style=None, show_size=False,
                     verbose=True, **kwargs):
    """Return the HTML code for an inspection slideshow of the training.

    This function must be used after the training was completed. Before the
    training prepare_training_inspection must have been called to create
    snapshots. After training one should call remove_inspection_residues.

    Note that the file into which the returned slideshow HTML is inserted must
    be in the snapshot_path.

    snapshot_path -- Path were the flow training snapshots are stored.
    x_samples, msg_samples -- Lists with the input data for the training trace.
    stop_messages -- The stop msg for the training trace.
    inspection_path -- Path were the slides will be stored. If None (default
        value) then the snapshot_path is used.
    tracer -- Instance of InspectionHTMLTracer, can be None for
        default class.
    debug -- If True (default is False) then any exception will be
        caught and the gathered data up to that point is returned in the
        normal way. This is useful for bimdp debugging.
    slide_style -- CSS code for the individual slides (when they are
        viewed as single HTML files), has no effect on the slideshow appearance.
    show_size -- Show the approximate memory footprint of all nodes.
    verbose -- If True (default value) a status message is printed for each
        loaded snapshot.
    **kwargs -- Additional arguments for flow.train can be specified
        as keyword arguments.
    """
    if not inspection_path:
        inspection_path = snapshot_path
    ## create CSS file for the slides
    if not slide_style:
        slide_style = standard_css()
    robust_write_file(path=inspection_path, filename=STANDARD_CSS_FILENAME,
                      content=slide_style)
    del slide_style
    ## create slides
    try:
        slide_filenames, slide_node_ids, index_table = \
            _trace_biflow_training(snapshot_path=snapshot_path,
                                   inspection_path=inspection_path,
                                   x_samples=x_samples,
                                   msg_samples=msg_samples,
                                   stop_messages=stop_messages,
                                   tracer=tracer,
                                   debug=debug,
                                   show_size=show_size,
                                   verbose=verbose,
                                   **kwargs )
        if not slide_filenames:
            err = ("No inspection slides were generated, probably because "
                   "there are no untrained nodes in the given flow.")
            raise EmptyTraceException(err)
    except TraceDebugException as debug_exception:
        slide_filenames, slide_node_ids, index_table = debug_exception.result
    if index_table is None:
        return None  # no snapshots were found
    # create slideshow
    slideshow = TrainHTMLSlideShow(filenames=slide_filenames,
                                   node_ids=slide_node_ids,
                                   index_table=index_table,
                                   delay=500, delay_delta=100, loop=False)
    return str(slideshow)

def show_training(flow, data_iterables, msg_iterables=None, stop_messages=None,
                  path=None, tracer=None,
                  debug=False,  show_size=False, open_browser=True,
                  **kwargs):
    """Perform both the flow training and the training inspection.

    The return value is the filename of the slideshow HTML file.
    This function must be used with the untrained flow (no previous call
    of Flow.train is required, the training happens here).

    This function is more convenient than inspect_training since it includes
    all required steps, but it is also less customizable. After everything
    is complete the inspection slideshow is opened in the browser.

    flow -- The untrained Flow or BiFlow. After this function has been called
        the flow will be fully trained.
    data_iterables, msg_iterables, stop_messages -- Same as for calling train
        on a flow.
    path -- Path were both the training snapshots and the inspection slides
        will be stored. If None (default value) a temporary directory will be
        used.
    tracer -- Instance of InspectionHTMLTracer, can be None for
        default class.
    debug -- Ignore exception during training and try to complete the slideshow
        (default value is False).
    show_size -- Show the approximate memory footprint of all nodes.
    open_browser -- If True (default value) then the slideshow file is
        automatically opened in a webbrowser. One can also use string value
        with the browser name (for webbrowser.get) to request a specific
        browser.
    **kwargs -- Additional arguments for flow.train can be specified
        as keyword arguments.
    """
    if path is None:
        path = tempfile.mkdtemp(prefix='MDP_')
    # get first part of data iterators as sample data for inspection
    # if data_iterables is an array, wrap it up in a list
    if isinstance(data_iterables, numx.ndarray):
        data_iterables = [[data_iterables]] * len(flow)
    x_samples = []
    for i, data_iterable in enumerate(data_iterables):
        if data_iterable is None:
            x_sample, new_data_iterable = None, None
        else:
            x_sample, new_data_iterable = first_iterable_elem(data_iterable)
        x_samples.append(x_sample)
        data_iterables[i] = new_data_iterable
    del x_sample
    if msg_iterables:
        msg_samples = []
        for i, msg_iterable in enumerate(msg_iterables):
            if msg_iterable is None:
                msg_sample, new_msg_iterable = None, None
            else:
                msg_sample, new_msg_iterable = first_iterable_elem(msg_iterable)
            msg_samples.append(msg_sample)
            msg_iterables[i] = new_msg_iterable
        del msg_sample
    else:
        msg_samples = None
    # store the data to disk to disk to save memory and safeguard against
    # any change made to the data during the training
    robust_pickle(path, "training_data_samples.pckl",
                  (x_samples, msg_samples, stop_messages))
    del x_samples
    del msg_samples
    # perform the training and gather snapshots
    prepare_training_inspection(flow=flow, path=path)
    try:
        if isinstance(flow, BiFlow):
            flow.train(data_iterables, msg_iterables, stop_messages, **kwargs)
        else:
            flow.train(data_iterables, **kwargs)
    except Exception:
        if debug:
            traceback.print_exc()
            print ("exception during training, " +
                   "inspecting up to failure point...")
            # create the last snapshot manually
            try:
                # if a normal mdp.Flow instance was given then this fails
                flow._bi_reset()
            except Exception:
                pass
            filename = (flow._snapshot_name_ + "_%d" % flow._snapshot_counter_
                        + PICKLE_EXT)
            robust_pickle(flow._snapshot_path_, filename, flow)
        else:
            raise
    remove_inspection_residues(flow)
    # reload data samples
    with open(os.path.join(path, "training_data_samples.pckl"), "rb") as sample_file:
        x_samples, msg_samples, stop_messages = pickle.load(sample_file)
    # create slideshow
    slideshow = inspect_training(snapshot_path=path,
                                 inspection_path=path,
                                 x_samples=x_samples,
                                 msg_samples=msg_samples,
                                 stop_messages=stop_messages,
                                 tracer=tracer,
                                 debug=debug, show_size=show_size,
                                 verbose=False)
    filename = os.path.join(path, "training_inspection.html")
    title = "Training Inspection"
    with open(filename, 'w') as html_file:
        html_file.write('<html>\n<head>\n<title>%s</title>\n' % title)
        html_file.write('<style type="text/css" media="screen">')
        html_file.write(standard_css())
        html_file.write('</style>\n</head>\n<body>\n')
        html_file.write('<h3>%s</h3>\n' % title)
        html_file.write(slideshow)
        html_file.write('</body>\n</html>')
    if open_browser:
        _open_custom_brower(open_browser, os.path.abspath(filename))
    return filename

def inspect_execution(flow, x, msg=None, target=None, path=None,
                      name=None, tracer=None, debug=False,
                      slide_style=None, show_size=False,
                      **kwargs):
    """Return the HTML code for an inspection slideshow of the execution
    and the return value of the execution (in a tuple).

    Note that the file into which the slideshow HTML is inserted must be in the
    snapshot_path.

    flow -- The flow for the execution.
    x, msg, target -- Data for the execution, msg and target can only be used
        for a BiFlow (default value is None).
    path -- Path were the slideshow will be stored, if None (default value)
        a temporary directory will be used.
    name -- Name string to be used for the slide files.
    tracer -- Instance of InspectionHTMLTracer, can be None for
        default class.
    debug -- If True (default is False) then any exception will be
        caught and the gathered data up to that point is returned in the
        normal way. This is useful for bimdp debugging.
    slide_style -- CSS code for the individual slides (when they are
        viewed as single HTML files), has no effect on the slideshow appearance.
    show_size -- Show the approximate memory footprint of all nodes.
    **kwargs -- Additional arguments for flow.execute can be specified
        as keyword arguments.
    """
    if path is None:
        path = tempfile.mkdtemp(prefix='MDP_')
    if not name:
        name = "execution_inspection"
    # create CSS file for the slides
    if not slide_style:
        slide_style = standard_css()
    robust_write_file(path=path, filename=STANDARD_CSS_FILENAME,
                      content=slide_style)
    del slide_style
    if not tracer:
        tracer = InspectionHTMLTracer()
        tracer._html_converter.flow_html_converter.show_size = show_size
    # create slides
    try:
        slide_filenames, slide_node_ids, section_ids, result = \
            tracer.trace_execution(path=path,
                                   trace_name=name,
                                   flow=flow,
                                   x=x, msg=msg, target=target,
                                   debug=debug,
                                   **kwargs)
    except TraceDebugException as debug_exception:
        if not debug_exception.result:
            return None
        traceback.print_exc()
        print ("exception during excecution, " +
               "create inspection up to failure point...")
        slide_filenames, slide_node_ids, section_ids = debug_exception.result
        result = None
    # create slideshow file
    if not slide_filenames:
        err = "For some reason no execution slides were generated."
        raise EmptyTraceException(err)
    if not section_ids:
        slideshow = ExecuteHTMLSlideShow(filenames=slide_filenames,
                                         node_ids=slide_node_ids,
                                         delay=500, delay_delta=100,
                                         loop=False)
    else:
        # after an exception the last section_id entry can be missing
        if len(section_ids) < len(slide_filenames):
            section_ids.append(section_ids[-1])
        slideshow = SectExecuteHTMLSlideShow(filenames=slide_filenames,
                                             node_ids=slide_node_ids,
                                             section_ids=section_ids,
                                             delay=500, delay_delta=100,
                                             loop=False)
    return str(slideshow), result

def show_execution(flow, x, msg=None, target=None, path=None, name=None,
                   tracer=None,
                   debug=False, show_size=False, open_browser=True,
                   **kwargs):
    """Write the inspection slideshow into an HTML file and open it in the
    browser.

    The return value is a tuple with the slideshow filename and the return
    value of the execution.

    flow -- The flow for the execution.
    x, msg, target -- Data for the execution, msg and target can only be used
        for a BiFlow (default value is None).
    path -- Path were the slideshow will be stored, if None (default value)
        a temporary directory will be used.
    name -- A name for the slideshow.
    tracer -- Instance of InspectionHTMLTracer, can be None for
        default class.
    debug -- If True (default is False) then any exception will be
        caught and the gathered data up to that point is returned in the
        normal way. This is useful for bimdp debugging.
    show_size -- Show the approximate memory footprint of all nodes.
    open_browser -- If True (default value) then the slideshow file is
        automatically opened in a webbrowser. One can also use string value
        with the browser name (for webbrowser.get) to request a specific
        browser.
    **kwargs -- Additional arguments for flow.execute can be specified
        as keyword arguments.
    """
    if path is None:
        path = tempfile.mkdtemp(prefix='MDP_')
    if not name:
        name = "execution_inspection"
        title = "Execution Inspection"
    else:
        title = "Execution Inspection: " + name
    filename = os.path.join(path, name + ".html")
    slideshow, result = inspect_execution(
                        flow=flow,
                        path=path,
                        x=x, msg=msg, target=target,
                        name=name,
                        tracer=tracer,
                        debug=debug,
                        show_size=show_size,
                        **kwargs)
    # inspect execution created the path if required, so no need to check here
    with open(filename, 'w') as html_file:
        html_file = open(filename, 'w')
        html_file.write('<html>\n<head>\n<title>%s</title>\n' % title)
        html_file.write('<style type="text/css" media="screen">')
        html_file.write(standard_css())
        html_file.write('</style>\n</head>\n<body>\n')
        html_file.write('<h3>%s</h3>\n' % title)
        html_file.write(slideshow)
        html_file.write('</body>\n</html>')
    if open_browser:
        _open_custom_brower(open_browser, os.path.abspath(filename))
    return filename, result



"""
BiMDP Flow class for flexible (bidirectional) data flow.

The central class is a BiFlow, which implements all the flow handling options
offered by the BiNode class (see binode.py for a description).
"""
from __future__ import print_function
from builtins import zip
from builtins import str
from builtins import range
from builtins import object

# NOTE: make sure that isinstance(str, target) is never used, so that in
#    principle any object could be used.

import itertools

import mdp
n = mdp.numx

from .binode import BiNode

# this target value tells the flow to abort and return the current values
EXIT_TARGET = "exit"


class NoneIterable(object):
    """Iterable for an infinite sequence of Nones."""

    def __iter__(self):
        while True:
            yield None


class BiFlowException(mdp.FlowException):
    """Exception for BiFlow problems."""
    pass


class MessageResultContainer(object):
    """Store and combine msg output chunks from a BiNode.

    It is for example used when the flow execution yields msg output, which has
    to be joined for the end result.
    """

    def __init__(self):
        """Initialize the internal storage variables."""
        self._msg_results = dict()  # all none array message results
        self._msg_array_results = dict()  # result dict for arrays

    def add_message(self, msg):
        """Add a single msg result to the combined results.

        msg must be either a dict of results or None. numpy arrays will be
        transformed to a single numpy array in the end. For all other types the
        addition operator will be used to combine results (i.e., lists will be
        appended, single integers will be summed over).
        """
        if msg:
            for key in msg:
                if type(msg[key]) is n.ndarray:
                    if key not in self._msg_array_results:
                        self._msg_array_results[key] = []
                    self._msg_array_results[key].append(msg[key])
                else:
                    if key not in self._msg_results:
                        self._msg_results[key] = msg[key]
                    else:
                        try:
                            self._msg_results[key] += msg[key]
                        except:
                            err = ("Could not combine final msg results "
                                   "in BiFlow.")
                            raise BiFlowException(err)

    def get_message(self):
        """Return the msg which combines all the msg results."""
        # move array results from _msg_array_results to _msg_results
        for key in self._msg_array_results:
            if key in self._msg_results:
                err = ("A key in the msg results is used with "
                       "different data types.")
                raise BiFlowException(err)
            else:
                self._msg_results[key] = n.concatenate(
                                                self._msg_array_results[key])
        return self._msg_results


class BiFlow(mdp.Flow):
    """BiMDP version of a flow, which supports jumps between nodes.

    This capabilities can be used by classes derived from BiNode.

    Normal nodes can also be used in this flow, the msg argument is skipped
    for these. Normal nodes can be also jump targets, but only when a relative
    target index is used (since they do not support node ids).
    """

    def __init__(self, flow, verbose=False, **kwargs):
        kwargs["crash_recovery"] = False
        super(BiFlow, self).__init__(flow=flow, verbose=verbose, **kwargs)

    ### Basic Methods from Flow. ###

    def train(self, data_iterables, msg_iterables=None,
              stop_messages=None):
        """Train the nodes in the flow.

        The nodes will be trained according to their place in the flow.

        data_iterables -- Sequence of iterables with the training data for each
            trainable node. Can also be a single array or None.
            Note that iterables yielding tuples for additonal node arguments
            (e.g. the class labels for an FDANode) are not supported in a
            BiFlow. Instead use the BiNode version of the node and provide
            the arguments in the message (via msg_iterables).
        msg_iterables -- Sequence of iterables with the msg training data
            for each trainable node.
        stop_messages -- Sequence of messages for stop_training.

        Note that the type and iterator length of the data iterables is taken
        as reference, so the message iterables are assumed to have the
        same length.
        """
        # Note: When this method is updated BiCheckpointFlow should be updated
        #    as well.
        self._bi_reset()  # normaly not required, just for safety
        data_iterables, msg_iterables = self._sanitize_training_iterables(
                                            data_iterables=data_iterables,
                                            msg_iterables=msg_iterables)
        if stop_messages is None:
            stop_messages = [None] * len(data_iterables)
        # train each Node successively
        for i_node in range(len(self.flow)):
            if self.verbose:
                print ("training node #%d (%s)" %
                       (i_node, str(self.flow[i_node])))
            self._train_node(data_iterables[i_node], i_node,
                             msg_iterables[i_node], stop_messages[i_node])
            if self.verbose:
                print("training finished")

    def _train_node(self, iterable, nodenr, msg_iterable=None,
                    stop_msg=None):
        """Train a particular node.

        nodenr -- index of the node to be trained
        msg_iterable -- optional msg data for the training
            Note that the msg is only passed to the Node if it is an instance
            of BiNode.
        stop_msg -- optional msg data for stop_training
            Note that the message is only passed to the Node if the msg is not
            None, so for a normal node the msg has to be None.

        Note: unlike the normal mdp.Flow we do no exception handling here.
        """
        if not self.flow[nodenr].is_trainable():
            return
        iterable, msg_iterable, _ = self._sanitize_iterables(iterable,
                                                             msg_iterable)
        while True:
            if not self.flow[nodenr].get_remaining_train_phase():
                break
            self._train_node_single_phase(iterable, nodenr,
                                          msg_iterable, stop_msg)
            self._bi_reset()

    def _train_node_single_phase(self, iterable, nodenr,
                                 msg_iterable, stop_msg=None):
        """Perform a single training phase for a given node.

        This method should be only called internally in BiFlow.
        """
        empty_iterator = True
        for (x, msg) in zip(iterable, msg_iterable):
            empty_iterator = False
            ## execute the flow until the nodes return value is right
            i_node = 0
            while True:
                result = self._execute_seq(x, msg, i_node=i_node,
                                           stop_at_node=nodenr)
                ## check the execution result, target should be True
                if (not isinstance(result, tuple)) or (len(result) != 3):
                    err = ("The Node to be trained was not reached " +
                           "during training, last result: " + str(result))
                    raise BiFlowException(err)
                elif result[2] is True:
                    x = result[0]
                    msg = result[1]
                else:
                    err = ("Target node not found in flow during " +
                           "training, last target value: " + str(result[2]))
                    raise BiFlowException(err)
                ## perform node training
                if isinstance(self.flow[nodenr], BiNode):
                    result = self.flow[nodenr].train(x, msg)
                    if result is None:
                        # training is done for this chunk
                        break
                else:
                    try:
                        self.flow[nodenr].train(x)
                    except TypeError:
                        # check if error is caused by additional node arguments
                        train_arg_keys = self._get_required_train_args(
                                                        self.flow[nodenr])
                        if len(train_arg_keys):
                            err = ("The node '%s' " % str(self.flow[nodenr]) +
                                   "requires additional training " +
                                   " arguments, which is not supported in a " +
                                   "BiFlow. Instead use the BiNode version " +
                                   "of the node and put the arguments in " +
                                   "the msg.")
                            raise BiFlowException(err)
                        else:
                            raise
                    break
                ## training execution continues, interpret result
                if not isinstance(result, tuple):
                    x = result
                    msg = None
                    target = None
                elif len(result) == 2:
                    x, msg = result
                    target = None
                elif len(result) == 3:
                    x, msg, target = result
                else:
                    err = ("Node produced invalid return value " +
                           "during training: " + str(result))
                    raise BiFlowException(err)
                i_node = self._target_to_index(target, nodenr)
            self._bi_reset()
        if empty_iterator:
            if self.flow[nodenr].get_current_train_phase() == 1:
                err_str = ("The training data iteration for node "
                           "no. %d could not be repeated for the "
                           "second training phase, you probably "
                           "provided an iterable instead of an "
                           "iterable." % (nodenr+1))
                raise BiFlowException(err_str)
            else:
                err = ("The training data iterable for node "
                       "no. %d is empty." % (nodenr+1))
                raise BiFlowException(err)
        ## stop_training part
        # unlike the normal mdp.Flow we always close the training
        # to perform the stop_training phase
        self._stop_training_hook()
        if stop_msg is None:
            result = self.flow[nodenr].stop_training()
        else:
            result = self.flow[nodenr].stop_training(stop_msg)
        if result is None:
            # the training phase ends here without an execute phase
            return
        # start an execution phase
        if not isinstance(result, tuple):
            x = result
            msg = None
            target = None
        elif len(result) == 2:
            x, msg = result
            target = None
        elif len(result) == 3:
            x, msg, target = result
            if target == EXIT_TARGET:
                return
        else:
            err = ("Node produced invalid return value " +
                   "for stop_training: " + str(result))
            raise BiFlowException(err)
        i_node = self._target_to_index(target, nodenr)
        result = self._execute_seq(x, msg, i_node=i_node)
        # check that we reached the end of flow or get EXIT_TARGET,
        # only complain if the target was not found
        if isinstance(result, tuple) and len(result) == 3:
            target = result[2]
            if target not in [1, -1, EXIT_TARGET]:
                err = ("Target node not found in flow during " +
                       "stop_training phase, last target value: " +
                       str(target))
                raise BiFlowException(err)

    def execute(self, iterable, msg_iterable=None, target_iterable=None):
        """Execute the flow and return (y, msg).

        Note that the returned msg can be an empty dict, but not None.

        iterable -- Can be an iterable or iterator for arrays, a single array
            or None. In the last two cases it is assumed that msg is a single
            message as well.
        msg_iterable -- Can be an iterable or iterator or a single message
            (but only if iterable is a single array or None).
        target_iterable -- Like msg_iterable, but for target.

        Note that the type and iteration length of iterable is taken as
        reference, so msg is assumed to have the same length.

        If msg results are found and if iteration is used then the BiFlow
        tries to join the msg results (and concatenate in the case of arrays).
        """
        self._bi_reset()  # normaly not required, just for safety
        iterable, msg_iterable, target_iterable = \
            self._sanitize_iterables(iterable, msg_iterable, target_iterable)
        y_results = None
        msg_results = MessageResultContainer()
        empty_iterator = True
        for (x, msg, target) in zip(iterable, msg_iterable,
                                               target_iterable):
            empty_iterator = False
            if not target:
                i_node = 0
            else:
                i_node = self._target_to_index(target)
            result = self._execute_seq(x=x, msg=msg, i_node=i_node)
            if not isinstance(result, tuple):
                y = result
                msg = None
            elif (len(result) == 2):
                y, msg = result
            elif (len(result) == 3) and (result[2] in [1, -1, EXIT_TARGET]):
                # target -1 is allowed for easier inverse handling
                y, msg = result[:2]
            elif len(result) == 3:
                err = ("Target node not found in flow during execute," +
                       " last result: " + str(result))
                raise BiFlowException(err)
            else:
                err = ("BiNode execution returned invalid result type: " +
                       result)
                raise BiFlowException(err)
            self._bi_reset()
            if msg:
                msg_results.add_message(msg)
            # check if all y have the same type and store it
            # note that the checks for msg are less restrictive
            if y is not None:
                if y_results is None:
                    y_results = [y]
                elif y_results is False:
                    err = "Some but not all y return values were None."
                    raise BiFlowException(err)
                else:
                    y_results.append(y)
            else:
                if y_results is None:
                    y_results = False
                else:
                    err = "Some but not all y return values were None."
                    raise BiFlowException(err)
        if empty_iterator:
            err = ("The execute data iterable is empty.")
            raise BiFlowException(err)
        # consolidate results
        if y_results:
            y_results = n.concatenate(y_results)
        result_msg = msg_results.get_message()
        return y_results, result_msg

    def __call__(self, iterable, msg_iterable=None):
        """Calling an instance is equivalent to call its 'execute' method."""
        return self.execute(iterable, msg_iterable=msg_iterable)

    ### New Methods for BiMDP. ###

    def _bi_reset(self):
        """Reset the nodes and internal flow variables."""
        for node in self.flow:
            if isinstance(node, BiNode):
                node.bi_reset()

    def _request_node_id(self, node_id):
        """Return first hit of _request_node_id on internal nodes.

        So _request_node_id is called for all nodes in the flow until a return
        value is not None. If no such node is found the return value is None.
        """
        for node in self.flow:
            if isinstance(node, BiNode):
                found_node = node._request_node_id(node_id)
                if found_node:
                    return found_node
        return None

    ## container special methods to support node_id

    def __getitem__(self, key):
        if isinstance(key, __builtins__['str']):
            item = self._request_node_id(key)
            if item is None:
                err = ("This biflow contains no node with with the id " +
                       str(key))
                raise KeyError(err)
            return item
        else:
            return super(BiFlow, self).__getitem__(key)

    def __setitem__(self, key, value):
        if isinstance(key, __builtins__['str']):
            err = "Setting nodes by node_id is not supported."
            raise BiFlowException(err)
        else:
            super(BiFlow, self).__setitem__(key, value)

    def __delitem__(self, key):
        if isinstance(key, __builtins__['str']):
            err = "Deleting nodes by node_id is not supported."
            raise BiFlowException(err)
        else:
            super(BiFlow, self).__delitem__(key)

    def __contains__(self, key):
        if isinstance(key, __builtins__['str']):
            if self._request_node_id(key) is not None:
                return True
            else:
                return False
        else:
            return super(BiFlow, self).__contains__(key)

    ### Flow Implementation Methods ###

    def _sanitize_training_iterables(self, data_iterables, msg_iterables):
        """Check and adjust the training iterable list."""
        if data_iterables is None:
            if msg_iterables is None:
                err = ("Both the training data and the training messages are "
                       "None.")
                raise BiFlowException(err)
            else:
                data_iterables = [None] * len(self.flow)
        elif isinstance(data_iterables, n.ndarray):
            data_iterables = [[data_iterables]] * len(self.flow)
            # the form of msg_iterables follows that of data_iterables
            msg_iterables = [[msg_iterables]] * len(data_iterables)
        else:
            data_iterables = self._train_check_iterables(data_iterables)
            if msg_iterables is None:
                msg_iterables = [None] * len(self.flow)
            else:
                msg_iterables = self._train_check_iterables(msg_iterables)
        return data_iterables, msg_iterables

    def _sanitize_iterables(self, iterable, msg_iterable, target_iterable=None):
        """Check and adjust a data, message and target iterable."""
        # TODO: maybe add additional checks
        if isinstance(iterable, n.ndarray):
            iterable = [iterable]
            msg_iterable = [msg_iterable]
            target_iterable = [target_iterable]
        elif iterable is None:
            if msg_iterable is None:
                err = "Both the data and the message iterable is None."
                raise BiFlowException(err)
            else:
                iterable = NoneIterable()
                if isinstance(msg_iterable, dict):
                    msg_iterable = [msg_iterable]
                    target_iterable = [target_iterable]
        else:
            if msg_iterable is None:
                msg_iterable = NoneIterable()
            if target_iterable is None:
                target_iterable = NoneIterable()
        return iterable, msg_iterable, target_iterable

    def _target_to_index(self, target, current_node=0):
        """Return the absolute node index of the target code.

        If the string id target node is not found in this flow then the string
        is returned without alteration.

        When a relative index is given it is translated to the absolute index
        and it is checked if it is in the allowed range.

        target -- Can be a string node id, a relative index or None (which
            is interpreted as 1).
        current_node -- If target is specified as a relative index then this
            node index is used to translate the target to the absolute node
            index (otherwise it has no effect).
        check_bounds -- If False then it is not checked wether the node index
            is in range(len(flow)).
        """
        if target == EXIT_TARGET:
            return EXIT_TARGET
        if target is None:
            target = 1
        if not isinstance(target, int):
            for i_node, node in enumerate(self.flow):
                if isinstance(node, BiNode) and node._request_node_id(target):
                    return i_node
            # no matching node was found
            return target
        else:
            absolute_index = current_node + target
            if absolute_index < 0:
                err = "Target int value references node at position < 0."
                raise BiFlowException(err)
            elif absolute_index >= len(self.flow):
                err = ("Target int value references a node"
                       " beyond the flow length (target " + str(target) +
                       ", current node " + str(current_node) + ").")
                raise BiFlowException(err)
            return absolute_index

    # TODO: update docstring for case when target is not found

    def _execute_seq(self, x, msg=None, i_node=0, stop_at_node=None):
        """Execute the whole flow as far as possible.

        i_node -- Can specify a node index where the excecution is supposed to
            start.
        stop_at_node -- Node index where the execution should stop. The input
            values for this node are returned in this case in the form
            (x, msg, target) with target being set to True.

        If the end of the flow is reached then the return value is y
        or (y, msg).
        If the an execution target node is not found then (x, msg, target) is
        returned (target values of 1 and -1 are also possible).
        If a normal Node (not derived from BiNode) is encountered then the
        current msg is simply carried forward around it.
        """
        ## this method is also used by other classes, like BiFlowNode
        while i_node != stop_at_node:
            if isinstance(self.flow[i_node], BiNode):
                result = self.flow[i_node].execute(x, msg)
                # check the type of the result
                if type(result) is not tuple:
                    x = result
                    msg = None
                    target = 1
                elif len(result) == 2:
                    x, msg = result
                    target = 1
                elif len(result) == 3:
                    x, msg, target = result
                else:
                    err = ("BiNode execution returned invalid result type: " +
                           result)
                    raise BiFlowException(err)
            else:
                # just a normal MDP node
                x = self.flow[i_node].execute(x)
                # note that the message is carried forward unchanged
                target = 1
            ## check if the target is in this flow, return otherwise
            if isinstance(target, int):
                i_node = i_node + target
                # values of +1 and -1 beyond this flow are allowed
                if i_node == len(self.flow):
                    if not msg:
                        return x
                    else:
                        return (x, msg)
                elif i_node == -1:
                    return x, msg, -1
            else:
                i_node = self._target_to_index(target, i_node)
                if not isinstance(i_node, int):
                    # target not found in this flow
                    # this is also the exit point when EXIT_TARGET is given
                    return x, msg, target
        # reached stop_at_node, signal this by returning target value True
        return x, msg, True


### Some useful flow classes. ###

class BiCheckpointFlow(BiFlow, mdp.CheckpointFlow):
    """Similar to normal checkpoint flow.

    The main difference is that even the last training phase of a
    node is already closed before the checkpoint function is called.
    """

    def train(self, data_iterables, checkpoints, msg_iterables=None,
              stop_messages=None):
        """Train the nodes in the flow.

        The nodes will be trained according to their place in the flow.

        Additionally calls the checkpoint function 'checkpoint[i]'
        when the training phase of node #i is over.
        A checkpoint function takes as its only argument the trained node.
        If the checkpoint function returns a dictionary, its content is
        added to the instance's dictionary.
        The class CheckpointFunction can be used to define user-supplied
        checkpoint functions.
        """
        self._bi_reset()  # normaly not required, just for safety
        data_iterables, msg_iterables = self._sanitize_training_iterables(
                                                 data_iterables=data_iterables,
                                                 msg_iterables=msg_iterables)
        if stop_messages is None:
            stop_messages = [None] * len(data_iterables)
        checkpoints = self._train_check_checkpoints(checkpoints)
        # train each Node successively
        for i_node in range(len(self.flow)):
            if self.verbose:
                print ("training node #%d (%s)" %
                       (i_node, str(self.flow[i_node])))
            self._train_node(data_iterables[i_node], i_node,
                             msg_iterables[i_node], stop_messages[i_node])
            if i_node <= len(checkpoints) and checkpoints[i_node] is not None:
                checkpoint_dict = checkpoints[i_node](self.flow[i_node])
                if dict:
                    self.__dict__.update(checkpoint_dict)
            if self.verbose:
                print("training finished")



from builtins import range
from builtins import object
from ._tools import *

import mdp.parallel as parallel
import mdp.hinet as hinet
n = numx

class TestParallelHinetNodes(object):
    """Tests for ParallelFlowNode."""

    def setup_method(self, method):
        if "parallel" in mdp.get_active_extensions():
            self.set_parallel = False
        else:
            mdp.activate_extension("parallel")
            self.set_parallel = True

    def teardown_method(self, method):
        if self.set_parallel:
            mdp.deactivate_extension("parallel")

    def test_flownode(self):
        """Test ParallelFlowNode."""
        flow = mdp.Flow([mdp.nodes.SFANode(output_dim=5),
                         mdp.nodes.PolynomialExpansionNode(degree=2),
                         mdp.nodes.SFANode(output_dim=3)])
        flownode = mdp.hinet.FlowNode(flow)
        x = n.random.random([100,50])
        chunksize = 25
        chunks = [x[i*chunksize : (i+1)*chunksize]
                    for i in range(len(x)//chunksize)]
        while flownode.get_remaining_train_phase() > 0:
            for chunk in chunks:
                forked_node = flownode.fork()
                forked_node.train(chunk)
                flownode.join(forked_node)
            flownode.stop_training()
        # test execution
        flownode.execute(x)
        
    def test_flownode_forksingle(self):
        """Test that ParallelFlowNode forks only the first training node."""
        flow = mdp.Flow([mdp.nodes.SFANode(output_dim=5),
                         mdp.nodes.PolynomialExpansionNode(degree=2),
                         mdp.nodes.SFANode(output_dim=3)])
        flownode = mdp.hinet.FlowNode(flow)
        forked_flownode = flownode.fork()
        assert flownode._flow[0] is not forked_flownode._flow[0]
        assert flownode._flow[1] is forked_flownode._flow[1]
        assert flownode._flow[2] is forked_flownode._flow[2]
        # Sabotage joining for the second SFANode, which should not be joined,
        # causing AttributeError: 'NoneType' ... when it is joined.
        flownode._flow[2]._cov_mtx = None
        flownode.join(forked_flownode)

    def test_parallelnet(self):
        """Test a simple parallel net with big data.

        Includes ParallelFlowNode, ParallelCloneLayer, ParallelSFANode
        and training via a ParallelFlow.
        """
        noisenode = mdp.nodes.NormalNoiseNode(input_dim=20*20,
                                              noise_args=(0,0.0001))
        sfa_node = mdp.nodes.SFANode(input_dim=20*20, output_dim=10)
        switchboard = hinet.Rectangular2dSwitchboard(in_channels_xy=100,
                                                     field_channels_xy=20,
                                                     field_spacing_xy=10)
        flownode = mdp.hinet.FlowNode(mdp.Flow([noisenode, sfa_node]))
        sfa_layer = mdp.hinet.CloneLayer(flownode,
                                                switchboard.output_channels)
        flow = parallel.ParallelFlow([switchboard, sfa_layer])
        data_iterables = [None,
                          [n.random.random((10, 100*100)) for _ in range(3)]]
        scheduler = parallel.Scheduler()
        flow.train(data_iterables, scheduler=scheduler)

    def test_layer(self):
        """Test Simple random test with three nodes."""
        node1 = mdp.nodes.SFANode(input_dim=10, output_dim=5)
        node2 = mdp.nodes.SFANode(input_dim=17, output_dim=3)
        node3 = mdp.nodes.SFANode(input_dim=3, output_dim=1)
        layer = mdp.hinet.Layer([node1, node2, node3])
        flow = parallel.ParallelFlow([layer])
        data_iterables = [[n.random.random((10, 30)) for _ in range(3)]]
        scheduler = parallel.Scheduler()
        flow.train(data_iterables, scheduler=scheduler)





###################################################################
### After changing this file, please copy it so that files
### ../../{mbp,bimdp}/test/test_seed.py are identical.
###################################################################

import mdp

_SEED = None

def _compare_with_seed(seed):
    global _SEED
    if _SEED is None:
        _SEED = seed
    return _SEED == seed

def test_seed():
    seed = mdp.numx_rand.get_state()[1][0]
    assert _compare_with_seed(seed), (_SEED, seed)
    mdp.numx_rand.seed(seed+1)

def test_seed_clone():
    # we need two identical functions to check that the seed
    # is reset at every call
    seed = mdp.numx_rand.get_state()[1][0]
    assert _compare_with_seed(seed), (_SEED, seed)
    mdp.numx_rand.seed(seed+1)

def test_seed_reset():
    # this function resets the global _SEED, so
    # that we can call the tests several times in
    # a row with different seeds without getting a failure
    global _SEED
    _SEED = None



"""General routines and base classes for Support Vector Machine classifiers.

TODO: Implement some scaling. Either by special Scaling Node or internally.
"""
from builtins import map
from builtins import zip
from builtins import object

import mdp
from mdp import ClassifierCumulator
from itertools import count

class _LabelNormalizer(object):
    """This class provides a transparent mapping from arbitrary labels
    to a set of well-defined integers.

    TODO: This could actually be a node.
    TODO: Needs more refinement. E.g. could automatically round labels to +1, -1
    """
    def __init__(self, labels, mode=None):
        if mode is None:
            mode = "id"
        if mode == "id":
            # don't do anything.
            self.normalize = self._id
            self.revert = self._id
            return

        self._mode = mode
        self._labels = set(labels)
        self._mapping = {}
        self._inverse = {}
        if mode == "dual":
            if len(self._labels) > 2:
                msg = "In dual mode only two labels can be given"
                raise mdp.NodeException(msg)
            t_label_norm = list(zip(self._labels, [1, -1]))
            self._set_label_dicts(t_label_norm)
        elif mode == "multi":
            # enumerate from zero to len
            t_label_norm = list(zip(self._labels, count()))
            self._set_label_dicts(t_label_norm)
        else:
            msg = "Remapping mode not known"
            raise mdp.NodeException(msg)

    def _set_label_dicts(self, t_label_norm):
        self._mapping = dict(t_label_norm)
        self._inverse = dict((norm, label) for label, norm in t_label_norm)

        # check that neither original nor normalised labels have occured more than once
        if not (len(self._mapping) == len(t_label_norm) == len(self._inverse)):
            msg = "Error in label normalisation."
            raise mdp.NodeException(msg)

    def normalize(self, labels):
        return list(map(self._mapping.get, labels))

    def revert(self, norm_labels):
        return list(map(self._inverse.get, norm_labels))

    def _id(self, labels):
        return labels


class _SVMClassifier(ClassifierCumulator):
    """Base class for the SVM classifier nodes."""

    def __init__(self, input_dim=None, output_dim=None, dtype=None):
        self.normalizer = None
        super(_SVMClassifier, self).__init__(input_dim=input_dim,
                                             output_dim=output_dim,
                                             dtype=dtype)

    @staticmethod
    def is_invertible():
        return False




"""
Module for HTML trace slideshows.

The individual slides are the HTML files generated via the trace_inspection
module (the body of the HTML files is extracted and makes up a slide).
"""
from builtins import range


import os

from mdp.utils import HTMLSlideShow, SectionHTMLSlideShow


class ExecuteHTMLSlideShow(HTMLSlideShow):

    def __init__(self, filenames, node_ids, delay=500, delay_delta=100,
                  loop=False, **kwargs):
        """Return the complete HTML code for the slideshow.

        filenames -- Sequence of strings, containing the path for each slide.
        node_ids -- Sequence of the active node ids for each slide.
        """
        kwargs.update(vars())
        # create a list of the possible node ids
        unique_node_ids = list(set(node_ids))
        kwargs["unique_node_ids"] = unique_node_ids
        del kwargs["self"]
        super(ExecuteHTMLSlideShow, self).__init__(**kwargs)

    _SLIDESHOW_CSS_FILENAME = "trace_slideshow.css"

    @classmethod
    def slideshow_css(cls):
        css_filename = os.path.join(os.path.split(__file__)[0],
                                    cls._SLIDESHOW_CSS_FILENAME)
        with open(css_filename, 'r') as css_file:
            css = css_file.read()
        return css

    js_loadslide_template = r'''

    // maps slide index to active node id
    var slide_node_ids = $node_ids;
    // list of all node ids that are available
    var unique_node_ids = $unique_node_ids;

    $<js_loadhtml_template>

    that.loadSlide = function () {
        loadPage(slideselect[current_slide].value);
    }

    // is called by loadPage after the loading has happened
    function makeNodesClickable() {
        var i;
        for (i = 0; i < unique_node_ids.length; i += 1) {
            try {
                document.getElementById(unique_node_ids[i]).
                    addEventListener("click", nodeClickCallback, false);
            }
            catch (e) {
                // means that the requested node is added in a later training
                // phase and is therefore not yet in the DOM
            }
        }
    }

    function nodeClickCallback() {
        // TODO: use event.srcElement for IE (event.target for W3C)
        var node_id = this.id;
        // search for next occurance of this node id
        var i;
        for (i = current_slide + 1; i < slide_node_ids.length; i += 1) {
            if (slide_node_ids[i] === node_id) {
                current_slide = i;
                that.updateSlide();
                return;
            }
        }
        // alert("Node is not reached after this slide.");
    }

'''

    js_loadhtml_template = r'''
    /**
     * Code to load the body content from HTMl files and inject it.
     * inspired by http://www.xul.fr/ajax/responseHTML-attribute.html
     */

    // Extract body content from html content.
    function getBody(content) {
        var lowContent = content.toLowerCase();  // eliminate case sensitivity
        // deal with attributes
        var i_start = lowContent.indexOf("<body");
        if (i_start === -1) {
            return "";
        }
        i_start = lowContent.indexOf(">", i_start);
        if (i_start === -1) {
            return "";
        }
        var i_end = lowContent.lastIndexOf("</body>");
        if (i_end === -1) {
            i_end = lowContent.lastIndexOf("</html>");
        }
        // if no HTML then just grab everything till end.
        if (i_end === -1) {
            i_end = content.length;
        }
        return content.slice(i_start + 1, i_end);
    }

    // Return a XMLHttpRequest object (browser independent).
    function getXHR()
    {
        var request = false;
            try {
                request = new ActiveXObject('Msxml2.XMLHTTP');
            }
            catch (err2) {
                try {
                    request = new ActiveXObject('Microsoft.XMLHTTP');
                }
                catch (err3) {
                    try {
                        request = new XMLHttpRequest();
                    }
                    catch (err1) {
                        request = false;
                    }
                }
            }
        return request;
    }

    // Load an HTML page and inject the content.
    function loadPage(url) {
        var target = document.getElementById("html_display");
        var xhr = getXHR();
        xhr.onreadystatechange = function() {
            if(xhr.readyState == 4) {
                target.innerHTML = getBody(xhr.responseText);
                makeNodesClickable();
            }
        }
        xhr.open("GET", url, true);
        xhr.send(null);
    }
    '''

    # Note: We do not use an id prefix, since there is only one slideshow.
    html_bottom_template = r'''
<div id="html_display"></div>
'''


class SectExecuteHTMLSlideShow(SectionHTMLSlideShow, ExecuteHTMLSlideShow):
    """Execute slideshow with support for sections."""
    pass


class TrainHTMLSlideShow(SectionHTMLSlideShow, ExecuteHTMLSlideShow):

    def __init__(self, filenames, node_ids, index_table, **kwargs):
        """Return the complete HTML code for the slideshow.

        filenames -- Sequence of strings, containing the path for each slide.
        node_ids -- Sequence of the active node ids for each slide.
        index_table -- Nested lists with the index data generated by
            inspect_biflow_training (last slide indexed by node, phase, train
            and stop).
        """
        slideshow_id = self._get_random_id()
        n_nodes = len(index_table)
        n_phases = max([len(phase_indices)
                        for phase_indices in index_table])
        # create the table and mapping between slide index and phase and node
        train_id = 0  # id indexing phase, node and train or stop
        start_index = 0  # first slide index for the current phase
        end_index = 0  # last slide index for the current phase
        section_ids = []
        train_table = [[None for _ in range(n_nodes + 1)]
                       for _ in range(n_phases + 1)]
        # create labels for table
        train_table[0] = ['&nbsp;'] + ['node %d' % (i+1)
                                       for i in range(n_nodes)]
        for i_phase in range(n_phases):
            train_table[i_phase+1][0] = 'phase %d' % (i_phase + 1)
        for i_node in range(n_nodes):
            for i_phase in range(len(index_table[i_node])):
                end_index = index_table[i_node][i_phase][0]
                # train link stuff
                html_string = ('<span class="inactive_section" ' +
                        'id="%s_section_id_%d" ' % (slideshow_id, train_id) +
                        'onClick="%s.setSlide(%d);">train</span>' %
                            (slideshow_id, start_index) +
                        '&nbsp;')
                section_ids += [train_id,] * (end_index - start_index + 1)
                train_id += 1
                # stop link stuff
                start_index = end_index + 1
                end_index = index_table[i_node][i_phase][1]
                if start_index > end_index:
                    # this can happen due to an exception during training
                    start_index = end_index
                else:
                    html_string += ('<span class="inactive_section" ' +
                        'id="%s_section_id_%d" ' % (slideshow_id, train_id) +
                        'onClick="%s.setSlide(%d);">stop</span>' %
                            (slideshow_id, start_index))
                train_table[i_phase+1][i_node+1] = html_string
                section_ids += [train_id,] * (end_index - start_index + 1)
                train_id += 1
                start_index = end_index + 1
        kwargs["control_table"] = train_table
        kwargs["section_ids"] = section_ids
        kwargs["filenames"] = filenames
        kwargs.update(vars())
        del kwargs["self"]
        del kwargs["index_table"]
        super(TrainHTMLSlideShow, self).__init__(**kwargs)

    html_top_template = r'''
<table class="slideshow">
${{
for row in control_table:
    self.write('<tr>\n')
    for cell in row:
        self.write('<td> %s </td>\n' % cell)
    self.write('</tr>\n')
}}
</table>
<br>
'''

    html_controls_template = r'''
${{super(SectionHTMLSlideShow, self).html_controls_template(vars())}}
'''



"""Test caching extension."""
from builtins import range
import tempfile
from ._tools import *

requires_joblib = skip_on_condition(
    "not mdp.config.has_joblib",
    "This test requires the 'joblib' module.")


_counter = 0
class _CounterNode(mdp.Node):
    def __init__(self):
        super(_CounterNode, self).__init__()

    def is_trainable(self):
        return False

    def _execute(self, x):
        """The execute method has the side effect of increasing
        a global counter by one."""
        global _counter
        _counter += 1
        return x

@requires_joblib
def test_caching_extension():
    """Test that the caching extension is working at the global level."""

    global _counter
    _counter = 0
    node = _CounterNode()

    # before decoration the global counter is incremented at every call
    k = 0
    for i in range(3):
        x = mdp.numx.array([[i]], dtype='d')
        for j in range(2):
            k += 1
            assert mdp.numx.all(node.execute(x) == x)
            assert _counter == k

    # reset counter
    _counter = 0
    # activate the extension
    cachedir = tempfile.mkdtemp(prefix='mdp-tmp-joblib-cache.',
                                dir=py.test.mdp_tempdirname)
    mdp.caching.activate_caching(cachedir=cachedir)
    assert mdp.get_active_extensions() == ['cache_execute']

    # after decoration the global counter is incremented for each new 'x'
    for i in range(3):
        x = mdp.numx.array([[i]], dtype='d')
        for _ in range(2):
            assert mdp.numx.all(node.execute(x) == x)
            assert _counter == i + 1

    # after deactivation
    mdp.caching.deactivate_caching()
    assert mdp.get_active_extensions() == []
    # reset counter
    _counter = 0

    k = 0
    for i in range(3):
        x = mdp.numx.array([[i]], dtype='d')
        for j in range(2):
            k += 1
            assert mdp.numx.all(node.execute(x) == x)
            assert _counter == k

@requires_joblib
def test_different_instances_same_content():
    global _counter
    x = mdp.numx.array([[100.]], dtype='d')

    cachedir = tempfile.mkdtemp(prefix='mdp-tmp-joblib-cache.',
                                dir=py.test.mdp_tempdirname)
    mdp.caching.activate_caching(cachedir=cachedir)
    node = _CounterNode()
    _counter = 0

    # add attribute to make instance unique
    node.attr = 'unique'

    # cache x
    node.execute(x)
    assert _counter == 1
    # should be cached now
    node.execute(x)
    assert _counter == 1

    # create new instance, make it also unique and check that
    # result is still cached
    _counter = 0
    node = _CounterNode()
    node.attr = 'unique and different'
    node.execute(x)
    assert _counter == 1

    mdp.caching.deactivate_caching()

@requires_joblib
def test_caching_context_manager():
    global _counter
    node = _CounterNode()
    _counter = 0

    assert mdp.get_active_extensions() == []
    cachedir = tempfile.mkdtemp(prefix='mdp-tmp-joblib-cache.',
                                dir=py.test.mdp_tempdirname)
    with mdp.caching.cache(cachedir=cachedir):
        assert mdp.get_active_extensions() == ['cache_execute']

        for i in range(3):
            x = mdp.numx.array([[i]], dtype='d')
            for _ in range(2):
                assert mdp.numx.all(node.execute(x) == x)
                assert _counter == i + 1
    assert mdp.get_active_extensions() == []

@requires_joblib
def test_class_caching():
    """Test that we can cache individual classes."""
    cached = mdp.nodes.PCANode()
    notcached = mdp.nodes.SFANode()
    with mdp.caching.cache(cache_classes=[mdp.nodes.PCANode]):
        assert cached.is_cached()
        assert not notcached.is_cached()

@requires_joblib
def test_class_caching_functionality():
    """Test that cached classes really cache."""
    global _counter
    x = mdp.numx.array([[210]], dtype='d')

    node = _CounterNode()

    # here _CounterNode is not cached
    _counter = 0
    with mdp.caching.cache(cache_classes=[mdp.nodes.PCANode]):
        node.execute(x)
        assert _counter == 1
        node.execute(x)
        assert _counter == 2

    # here _CounterNode is cached
    _counter = 0
    with mdp.caching.cache(cache_classes=[_CounterNode]):
        node.execute(x)
        assert _counter == 1
        node.execute(x)
        assert _counter == 1

@requires_joblib
def test_instance_caching():
    """Test that we can cache individual instances."""
    cached = mdp.nodes.PCANode()
    notcached = mdp.nodes.PCANode()
    with mdp.caching.cache(cache_instances=[cached]):
        assert cached.is_cached()
        assert not notcached.is_cached()

@requires_joblib
def test_instance_caching_functionality():
    """Test that cached instances really cache."""
    global _counter
    x = mdp.numx.array([[130]], dtype='d')

    node = _CounterNode()
    othernode = _CounterNode()

    # here _CounterNode is not cached
    _counter = 0
    with mdp.caching.cache(cache_instances=[othernode]):
        node.execute(x)
        assert _counter == 1
        node.execute(x)
        assert _counter == 2

    # here _CounterNode is cached
    _counter = 0
    with mdp.caching.cache(cache_instances=[node]):
        node.execute(x)
        assert _counter == 1
        node.execute(x)
        assert _counter == 1

@requires_joblib
def test_preexecution_problem():
    """Test that automatic setting of e.g. input_dim does not stop
    the caching extension from caching on the first run."""
    global _counter
    x = mdp.numx.array([[102.]])

    node = _CounterNode()

    # here _CounterNode is cached
    _counter = 0
    with mdp.caching.cache():
        # on the first execution, input_dim and dtype are set ...
        node.execute(x)
        assert _counter == 1
        # ... yet the result is cached
        node.execute(x)
        assert _counter == 1

@requires_joblib
def test_switch_cache():
    """Test changing cache directory while extension is active."""
    global _counter

    dir1 = tempfile.mkdtemp(prefix='mdp-tmp-joblib-cache.',
                            dir=py.test.mdp_tempdirname)
    dir2 = tempfile.mkdtemp(prefix='mdp-tmp-joblib-cache.',
                            dir=py.test.mdp_tempdirname)
    x = mdp.numx.array([[10]], dtype='d')

    mdp.caching.activate_caching(cachedir=dir1)

    node = _CounterNode()
    _counter = 0
    node.execute(x)
    assert _counter == 1
    node.execute(x)
    assert _counter == 1

    # now change path
    mdp.caching.set_cachedir(cachedir=dir2)
    node.execute(x)
    assert _counter == 2
    node.execute(x)
    assert _counter == 2

    mdp.caching.deactivate_caching()


@requires_joblib
def test_execute_magic():
    """Test calling execute with magic while caching."""
    x = mdp.numx_rand.rand(100, 10)

    node = mdp.nodes.PCANode()

    with mdp.caching.cache():
        y = node(x)
        y2 = node(x)
        assert_array_equal(y, y2)



from __future__ import division
from builtins import range
from past.utils import old_div
from ._tools import *

def _uniform(min_, max_, dims):
    return uniform(dims)*(max_-min_)+min_

def test_NeuralGasNode():
    ### test 1D distribution in a 10D space
    # line coefficients
    dim = 10
    npoints = 1000
    const = _uniform(-100,100,[dim])
    dir = _uniform(-1,1,[dim])
    dir /= utils.norm2(dir)
    x = _uniform(-1,1,[npoints])
    data = numx.outer(x, dir)+const
    # train the ng network
    num_nodes = 10
    ng = mdp.nodes.NeuralGasNode(start_poss=[data[n,:]
                                             for n in range(num_nodes)],
                                 max_epochs=10)
    ng.train(data)
    ng.stop_training()
    # control that the nodes in the graph lie on the line
    poss = ng.get_nodes_position()-const
    norms = numx.sqrt(numx.sum(poss*poss, axis=1))
    poss = (old_div(poss.T,norms)).T
    assert max(numx.minimum(numx.sum(abs(poss-dir),axis=1),
                            numx.sum(abs(poss+dir),axis=1))) < 1e-7, \
           'At least one node of the graph does lies out of the line.'
    # check that the graph is linear (no additional branches)
    # get a topological sort of the graph
    topolist = ng.graph.topological_sort()
    deg = numx.asarray([n.degree() for n in topolist])
    idx = deg.argsort()
    deg = deg[idx]
    assert_equal(deg[:2],[1,1])
    assert_array_equal(deg[2:], [2 for i in range(len(deg)-2)])
    # check the distribution of the nodes' position is uniform
    # this node is at one of the extrema of the graph
    x0 = numx.outer(numx.amin(x, axis=0), dir)+const
    x1 = numx.outer(numx.amax(x, axis=0), dir)+const
    linelen = utils.norm2(x0-x1)
    # this is the mean distance the node should have
    dist = old_div(linelen, poss.shape[0])
    # sort the node, depth first
    nodes = ng.graph.undirected_dfs(topolist[idx[0]])
    poss = numx.array([n.data.pos for n in nodes])
    dists = numx.sqrt(numx.sum((poss[:-1,:]-poss[1:,:])**2, axis=1))
    assert_almost_equal(dist, mean(dists), 1)

def test_NeuralGasNode_nearest_neighbor():
    # test the nearest_neighbor function
    start_poss = [numx.asarray([2.,0]), numx.asarray([-2.,0])]
    ng = mdp.nodes.NeuralGasNode(start_poss=start_poss, max_epochs=4)
    x = numx.asarray([[2.,0]])
    ng.train(x)
    nodes, dists = ng.nearest_neighbor(numx.asarray([[3.,0]]))
    assert_almost_equal(dists[0], 1., 7)
    assert_almost_equal(nodes[0].data.pos, numx.asarray([2., 0.]), 7)



# Modular toolkit for Data Processing (MDP)
"""\
**The Modular toolkit for Data Processing (MDP)** package is a library
of widely used data processing algorithms, and the possibility to
combine them together to form pipelines for building more complex
data processing software.

MDP has been designed to be used as-is and as a framework for
scientific data processing development.

From the user's perspective, MDP consists of a collection of *units*,
which process data. For example, these include algorithms for
supervised and unsupervised learning, principal and independent
components analysis and classification.

These units can be chained into data processing flows, to create
pipelines as well as more complex feed-forward network
architectures. Given a set of input data, MDP takes care of training
and executing all nodes in the network in the correct order and
passing intermediate data between the nodes. This allows the user to
specify complex algorithms as a series of simpler data processing
steps.

The number of available algorithms is steadily increasing and includes
signal processing methods (Principal Component Analysis, Independent
Component Analysis, Slow Feature Analysis), manifold learning methods
([Hessian] Locally Linear Embedding), several classifiers,
probabilistic methods (Factor Analysis, RBM), data pre-processing
methods, and many others.

Particular care has been taken to make computations efficient in terms
of speed and memory. To reduce the memory footprint, it is possible to
perform learning using batches of data. For large data-sets, it is
also possible to specify that MDP should use single precision floating
point numbers rather than double precision ones. Finally, calculations
can be parallelised using the ``parallel`` subpackage, which offers a
parallel implementation of the basic nodes and flows.

From the developer's perspective, MDP is a framework that makes the
implementation of new supervised and unsupervised learning algorithms
easy and straightforward. The basic class, ``Node``, takes care of tedious
tasks like numerical type and dimensionality checking, leaving the
developer free to concentrate on the implementation of the learning
and execution phases. Because of the common interface, the node then
automatically integrates with the rest of the library and can be used
in a network together with other nodes.

A node can have multiple training phases and even an undetermined
number of phases. Multiple training phases mean that the training data
is presented multiple times to the same node. This allows the
implementation of algorithms that need to collect some statistics on
the whole input before proceeding with the actual training, and others
that need to iterate over a training phase until a convergence
criterion is satisfied. It is possible to train each phase using
chunks of input data if the chunks are given as an iterable. Moreover,
crash recovery can be optionally enabled, which will save the state of
the flow in case of a failure for later inspection.

MDP is distributed under the open source BSD license. It has been
written in the context of theoretical research in neuroscience, but it
has been designed to be helpful in any context where trainable data
processing algorithms are used. Its simplicity on the user's side, the
variety of readily available algorithms, and the reusability of the
implemented nodes also make it a useful educational tool.

http://mdp-toolkit.sourceforge.net
"""
__docformat__ = "restructuredtext en"

# The descriptions strings below are parsed with a regexp in setup.py.
# Don't do anything fancy, keep strings triple quoted and verify that
# the get_*_description functions continue to work.

# __short_description__ must be one line, 200 characters maximum.
# C.f. http://docs.python.org/distutils/setupscript.html?highlight=description#meta-data
__short_description__ = """\
MDP is a Python library for building complex data processing software \
by combining widely used machine learning algorithms into pipelines \
and networks."""

__medium_description__ ="""\
**Modular toolkit for Data Processing (MDP)** is a Python data processing framework.

From the user's perspective, MDP is a collection of supervised and unsupervised
learning algorithms and other data processing units that can be combined into
data processing sequences and more complex feed-forward network architectures.

From the scientific developer's perspective, MDP is a modular framework, which
can easily be expanded. The implementation of new algorithms is easy and
intuitive. The new implemented units are then automatically integrated with
the rest of the library.

The base of available algorithms is steadily increasing and includes
signal processing methods (Principal Component Analysis,
Independent Component Analysis, Slow Feature Analysis),
manifold learning methods ([Hessian] Locally Linear Embedding),
several classifiers, probabilistic methods (Factor Analysis, RBM),
data pre-processing methods, and many others.
"""

class MDPException(Exception):
    """Base class for exceptions in MDP."""
    pass

class MDPWarning(UserWarning):
    """Base class for warnings in MDP."""
    pass

class MDPDeprecationWarning(DeprecationWarning, MDPWarning):
    """Warn about deprecated MDP API."""
    pass


from . import configuration

__version__ = '3.5'
__revision__ = configuration.get_git_revision()
__authors__ = 'MDP Developers'
__copyright__ = '(c) 2003-2016 mdp-toolkit-devel@lists.sourceforge.net'
__license__ = 'BSD License, see COPYRIGHT'
__contact__ = 'mdp-toolkit-users@lists.sourceforge.net'
__homepage__ = 'http://mdp-toolkit.sourceforge.net'

configuration.set_configuration()

config = configuration.config
(numx_description, numx, numx_linalg, numx_fft,
 numx_rand, numx_version) = configuration.get_numx()

# import the utils module (used by other modules)
from . import utils
# set symeig
utils.symeig = configuration.get_symeig(numx_linalg)

# import exceptions from nodes and flows
from .signal_node import (NodeException, InconsistentDimException,
                         TrainingException,
                         TrainingFinishedException, IsNotTrainableException,
                         IsNotInvertibleException)
from .linear_flows import CrashRecoveryException, FlowException, FlowExceptionCR

# import base nodes and flow classes
from .signal_node import (NodeMetaclass, Node, PreserveDimNode,
                         Cumulator, VariadicCumulator)

from .linear_flows import (Flow, CheckpointFlow,
                          CheckpointFunction, CheckpointSaveFunction)

# import helper functions:
from .helper_funcs import pca, fastica

# import extension mechanism
from .extension import (ExtensionException, extension_method,
                       extension_setup, extension_teardown,
                       ExtensionNodeMetaclass,
                       ExtensionNode, get_extensions,
                       get_active_extensions, with_extension,
                       activate_extension, deactivate_extension,
                       activate_extensions, deactivate_extensions,
                       extension)

# import classifier node
from .classifier_node import (ClassifierNode, ClassifierCumulator)

# import our modules
from . import nodes
from . import hinet
from . import parallel
from .test import test

# explicitly set __all__, mainly needed for epydoc
__all__ = ['config',
           'CheckpointFlow',
           'CheckpointFunction',
           'CheckpointSaveFunction',
           'ClassifierCumulator',
           'ClassifierNode',
           'CrashRecoveryException',
           'Cumulator',
           'ExtensionNode',
           'ExtensionNodeMetaclass',
           'Flow',
           'FlowException',
           'FlowExceptionCR',
           'IsNotInvertibleException',
           'IsNotTrainableException',
           'MDPException',
           'MDPWarning',
           'Node',
           'NodeException',
           'TrainingException',
           'TrainingFinishedException',
           'VariadicCumulator',
           'activate_extension',
           'activate_extensions',
           'deactivate_extension',
           'deactivate_extensions',
           'extension',
           'extension_method',
           'extension_setup',
           'extension_teardown',
           'get_extensions',
           'graph',
           'hinet',
           'nodes',
           'parallel',
           'pca',
           'fastica',
           'utils',
           'with_extension',
           ]

if config.has_joblib:
    from . import caching
    __all__ += ['caching']

utils.fixup_namespace(__name__, __all__,
                      ('signal_node',
                       'linear_flows',
                       'helper_funcs',
                       'classifier_node',
                       'configuration',
                       'repo_revision',
                       'extension',
                       ),('extension',
                          'configuration'))



from future import standard_library
standard_library.install_aliases()
from builtins import range

import tempfile
import pickle
import mdp
from ._tools import BogusMultiNode, BogusNodeTrainable
import py.test

uniform = mdp.numx_rand.random
MAT_DIM = (500,5)

def test_Node_copy():
    test_list = [1,2,3]
    generic_node = mdp.Node()
    generic_node.dummy_attr = test_list
    copy_node = generic_node.copy()
    assert generic_node.dummy_attr == copy_node.dummy_attr,\
           'Node copy method did not work'
    copy_node.dummy_attr[0] = 10
    assert generic_node.dummy_attr != copy_node.dummy_attr,\
           'Node copy method did not work'

def test_Node_copy_with_arrays_and_subnodes():
    node = mdp.Node()
    node.node = mdp.Node()
    node.node.x = mdp.numx.zeros((2,2))
    node2 = node.copy()
    assert hasattr(node2, 'node')
    assert mdp.numx.all(node2.node.x == node.node.x) 

def test_Node_copy_with_lambdas():
    generic_node = mdp.Node()
    generic_node.lambda_function = lambda: 1
    generic_node.copy()
    
def test_Node_save():
    test_list = [1,2,3]
    generic_node = mdp.Node()
    generic_node.dummy_attr = test_list
    # test string save
    copy_node_pic = generic_node.save(None)
    copy_node = pickle.loads(copy_node_pic)
    assert generic_node.dummy_attr == copy_node.dummy_attr,\
           'Node save (string) method did not work'
    copy_node.dummy_attr[0] = 10
    assert generic_node.dummy_attr != copy_node.dummy_attr,\
           'Node save (string) method did not work'
    # test file save
    dummy_file = tempfile.mktemp(prefix='MDP_', suffix=".pic",
                                 dir=py.test.mdp_tempdirname)
    generic_node.save(dummy_file, protocol=1)
    dummy_file = open(dummy_file, 'rb')
    copy_node = pickle.load(dummy_file)
    assert generic_node.dummy_attr == copy_node.dummy_attr,\
           'Node save (file) method did not work'
    copy_node.dummy_attr[0] = 10
    assert generic_node.dummy_attr != copy_node.dummy_attr,\
           'Node save (file) method did not work'

def test_Node_multiple_training_phases():
    x = uniform(size=MAT_DIM)
    node = BogusMultiNode()
    phases = node.get_remaining_train_phase()
    for i in range(phases):
        assert node.get_current_train_phase() == i
        assert not node._train_phase_started
        node.train(x)
        assert node._train_phase_started
        node.stop_training()

    assert not node.is_training()

def test_Node_execution_without_training():
    x = uniform(size=MAT_DIM)
    # try execution without training: single train phase
    node = BogusNodeTrainable()
    node.execute(x)
    assert hasattr(node, 'bogus_attr')
    # multiple train phases
    node = BogusMultiNode()
    node.execute(x)
    assert node.visited == [1, 2, 3, 4]
